/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

#ifndef FILE_MULTI_ANALYZER_PROCESS
#define FILE_MULTI_ANALYZER_PROCESS

#include "MultiAnalyzerBase.h"
#include "Trading.h"
#include "Learning.h"
#include "LineProcessed.h"
#include "LinesProcessor.h"
#include "Params.h"
#include "../../Interoperability/DateTime.h"
#include "PastTradesAnalyzer.h"
#include "Rules.h"
#include "DistancesProvider.h"
#include "../LibrariesInteroperability.h"
#include "../LibrariesLogger.h"
#include "OpenTrades.h"
#include "Bool.h"

//+------------------------------------------------------------------+
class CMultiAnalyzerProcess : public CMultiAnalyzerBase{//Does analysis of Symbols on live data
    private:
        int m_SearchingProfit;
     
        CLearning *m_LearningBuys;
        CLearning *m_LearningSells;
    
        CParams *m_Params;
		CPastTradesAnalyzer *m_PastTradesAnalyzer;
		CLearning *m_PastTradesLearn;

		CDistancesProvider *m_DistancesProvider;

		COpenTrades *m_OpenTrades;

        double m_MaximumLoss;
        double m_MinimumProfit;
    
		CBool *m_ReverseTradesAndLotSize;

        void ConvertToLearnVector(CLineProcessed &line, CVector &vec, double upperLimit, double lowerLimit)//Converts from CLineProcessed to LernVector
        {
             vec.Init(line.Points.Size() + 1 + 1, 1);
             for(int i=0;i<line.Points.Size();i++)
             {
                 vec.Input.SetValue(i, line.Points.GetValue(i));
             }
             vec.Input.SetValue(line.Points.Size(), upperLimit);
             vec.Input.SetValue(line.Points.Size() + 1, lowerLimit);
        }
    
        bool ProceedWithBuy(IString &symbol, IDateTime &dateOfMinimumMaximum, IDateTime &dateOfEstablishingMinimumMaximum, double upperLimit, double lowerLimit)//Determines if to Proceed with Buy operation for given symbol
        {
            if(!CConfig::UseSellBuyNeuralNetworks)
            {
                return true;
            }
            if((*m_LearningBuys).IsValid())
            {
                CLineProcessed line;
                CVector vec;
            
                CLinesProcessor::GenerateDifferentialLine(symbol, dateOfMinimumMaximum, dateOfEstablishingMinimumMaximum, line);
                ConvertToLearnVector(line, vec, upperLimit, lowerLimit);
                
                return (*m_LearningBuys).IsPositive(vec);
            }
            else
            {
                return false;
            }
        }

        bool ProceedWithSell(IString &symbol, IDateTime &dateOfMinimumMaximum, IDateTime &dateOfEstablishingMinimumMaximum, double upperLimit, double lowerLimit)//Determines if to Proceed with Sell Operation for given Symbol
        {
            if(!CConfig::UseSellBuyNeuralNetworks)
            {
                return true;
            }
            if((*m_LearningSells).IsValid())
            {
                CLineProcessed line;
                CVector vec;
            
                CLinesProcessor::GenerateDifferentialLine(symbol, dateOfMinimumMaximum, dateOfEstablishingMinimumMaximum, line);
                ConvertToLearnVector(line, vec, upperLimit, lowerLimit);
                
                return (*m_LearningSells).IsPositive(vec);
            }
            else
            {
                return false;
            }
        }
        
		bool ProceedWithOperation(bool Buy, IString &symbol, IDateTime &dateOfMinimumMaximum, IDateTime &dateOfEstablishingMinimumMaximum, double upperLimit, double lowerLimit)//Determines if to Proceed with operation for given Symbol
		{
			if (Buy)
			{
				return ProceedWithBuy(symbol, dateOfMinimumMaximum, dateOfEstablishingMinimumMaximum, upperLimit, lowerLimit);
			}
			else
			{
				return ProceedWithSell(symbol, dateOfMinimumMaximum, dateOfEstablishingMinimumMaximum, upperLimit, lowerLimit);
			}
		}

        bool ProceedWithTrade(IString &symbol, int symbolNr, bool Symbol, bool SymbolBuy, IDateTime &dateOfMinimumMaximum, IDateTime &dateOfEstablishingMinimumMaximum, double upperLimit, double lowerLimit)//Determines if to proceed with Trade
        {
            if(!CConfig::UseIndividualNeuralNetworks)
            {
                return true;
            }
            if((*(*m_Pairs).ReturnLearningObject(symbolNr, Symbol, SymbolBuy)).IsValid())
            {
                CLineProcessed line;
                CVector vec;
            
                CLinesProcessor::GenerateDifferentialLine(symbol, dateOfMinimumMaximum, dateOfEstablishingMinimumMaximum, line);
                ConvertToLearnVector(line, vec, upperLimit, lowerLimit);
                
                return (*(*m_Pairs).ReturnLearningObject(symbolNr, Symbol, SymbolBuy)).IsPositive(vec);
            }
            else
            {
                return false;
            }
        }
    
        bool ReverseTrades(bool reverse)//If true Reverses Trades
        {
            bool res = false;
            
            if(reverse)
            {
                res = !res;
            }
            return res;
        }
        
		bool CheckTrendReverseTrades()//Determine if to apply Reverse Trade logic based on trend analysis
		{
			if (CConfig::UseTrendFromLogs && (*m_PastTradesAnalyzer).IsValid())
			{
				CVector vec;

				if ((*m_Params).FillVector(vec, (*m_PastTradesAnalyzer).BestInputsAmount()))
				{
					if (CConfig::PastTrendsUseNeuralNetwork && (*m_PastTradesLearn).IsValid())
					{
						if ((*m_PastTradesLearn).IsPositive(vec))
						{
							return false;
						}
						else
						{
							return true;
						}
					}
					else
					{
						if ((*m_PastTradesAnalyzer).DeterminePrediction(vec))
						{
							return false;
						}
						else
						{
							return true;
						}
					}
				}
				else
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}

		bool CheckParamsVector()//Check if there is enought data to fill vector
		{
			if (CConfig::UseTrendFromLogs)
			{
				CVector vec;

				if ((*m_Params).FillVector(vec, (*m_PastTradesAnalyzer).BestInputsAmount()))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				return true;
			}
		}

		bool AfterTrendReverseTrades()//Check if to perform reverse trades logic based on after trend analysis
		{
			if (CConfig::UseAfterTrend)
			{
				(*m_Params).WereAfterTradesSuccesfull();
			}
			return (*m_Params).AfterTrendReverseTrades;
		}

		bool IsRuleHit(bool reverseTrades)//Detrmines if conditions were met to use any of the specified rules
		{
			CRules::ReadVaribles();
			double speed;

			if ((*m_Params).GetSpeed(CConfig::LastHours, speed))
			{
				bool hit = CRules::HitsRule(speed, reverseTrades);

				GetLogger().LogSpecial(IString("Hit : ") + GetInteroperability().BoolToString(hit) + IString(", Speed = ") + GetInteroperability().DoubleToString(speed) + IString(", ReverseTrades : ") + GetInteroperability().BoolToString(reverseTrades));
				return hit;
			}
			else
			{
				GetLogger().LogSpecial(IString("IsRuleHit : not valid."));
				return false;
			}
		}

		bool CheckDistance(bool Symbol, bool Buy, int SymbolNr, bool &MethodFReverseTrades)//Determines if to proceed based on calculation of Distance of current situation to past situations
		{
			if (CConfig::UseCalculateDistance)
			{
				if ((*m_DistancesProvider).IsValid())
				{
					bool proceed;
					MethodFReverseTrades = false;
			
					FillPresentMinimasMaximas();
					proceed = (*m_DistancesProvider).Calculate(Symbol, Buy, SymbolNr, MethodFReverseTrades);

					GetLogger().LogSpecial(IString("Check Distance : ") + GetInteroperability().BoolToString(proceed));
					return proceed;
				}
				else
				{
					MethodFReverseTrades = false;
					GetLogger().LogSpecial(IString("Check Distance : Distance Provider Not Valid."));
					return false;
				}
			}
			else
			{
				return false;
			}
		}

		void FillPresentMinimasMaximas()//Filss Distance Provider with information about current situation
		{
			IDateTime date = IDateTime::TimeCurrent();

			for (int i = 0; i < (*m_Pairs).AmountCurrencyPairsToBuy(); i++)
			{
				if ((*(m_AnalyzersForToBuy.GetPointerToValue(i))).ReturnLastMinimumMaximumDate().IsTimeValid())
				{
					(*m_DistancesProvider).CurrBuyTimes.SetValue(i, IDateTime::SecondsDiff((*(m_AnalyzersForToBuy.GetPointerToValue(i))).ReturnLastMinimumMaximumDate(), date));
					(*m_DistancesProvider).CurrBuyMinimasMaximas.SetValue(i, !((*(m_AnalyzersForToBuy.GetPointerToValue(i))).IsSearchingForMaximum()));
				}
				else
				{
					(*m_DistancesProvider).CurrBuyTimes.SetValue(i, -1.0);
				}
			}
			for (int i = 0; i < (*m_Pairs).AmountCurrencyPairsToSell(); i++)
			{
				if ((*(m_AnalyzersForToSell.GetPointerToValue(i))).ReturnLastMinimumMaximumDate().IsTimeValid())
				{
					(*m_DistancesProvider).CurrSellTimes.SetValue(i, IDateTime::SecondsDiff((*(m_AnalyzersForToSell.GetPointerToValue(i))).ReturnLastMinimumMaximumDate(), date));
					(*m_DistancesProvider).CurrSellMinimasMaximas.SetValue(i, !((*(m_AnalyzersForToSell.GetPointerToValue(i))).IsSearchingForMaximum()));
				}
				else
				{
					(*m_DistancesProvider).CurrSellTimes.SetValue(i, -1.0);
				}
			}
			for (int i = 0; i < (*m_Pairs).AmountOtherSymbolsToBuy(); i++)
			{
				if ((*(m_AnalyzersOtherForToBuy.GetPointerToValue(i))).ReturnLastMinimumMaximumDate().IsTimeValid())
				{
					(*m_DistancesProvider).OtherBuyTimes.SetValue(i, IDateTime::SecondsDiff((*(m_AnalyzersOtherForToBuy.GetPointerToValue(i))).ReturnLastMinimumMaximumDate(), date));
					(*m_DistancesProvider).OtherBuyMinimasMaximas.SetValue(i, !((*(m_AnalyzersOtherForToBuy.GetPointerToValue(i))).IsSearchingForMaximum()));
				}
				else
				{
					(*m_DistancesProvider).OtherBuyTimes.SetValue(i, -1.0);
				}
			}
			for (int i = 0; i < (*m_Pairs).AmountOtherSymbolsToSell(); i++)
			{
				if ((*(m_AnalyzersOtherForToSell.GetPointerToValue(i))).ReturnLastMinimumMaximumDate().IsTimeValid())
				{
					(*m_DistancesProvider).OtherSellTimes.SetValue(i, IDateTime::SecondsDiff((*(m_AnalyzersOtherForToSell.GetPointerToValue(i))).ReturnLastMinimumMaximumDate(), date));
					(*m_DistancesProvider).OtherSellMinimasMaximas.SetValue(i, !((*(m_AnalyzersOtherForToSell.GetPointerToValue(i))).IsSearchingForMaximum()));
				}
				else
				{
					(*m_DistancesProvider).OtherSellTimes.SetValue(i, -1.0);
				}
			}
		}

    public:

		void AssignDistancesProvider(CDistancesProvider &dp)//Assigns DistanceProvider object
		{
			m_DistancesProvider = &dp;
		}

        void ZeroSearchingProfit()//Zeros profit
        {
            m_SearchingProfit = 0;        
        }
    
        void AssignMaximumMinimumLossProfit(double loss, double profit)//Assigns loss, profit
        {
            m_MaximumLoss = loss;
            m_MinimumProfit = profit;
        }
    
        void AssignNetworks(CLearning &buys, CLearning &sells)//Assigns trained Neral Networks to object
        {
            m_LearningBuys = &buys;
            m_LearningSells = &sells;
        }

		void SetPastTradesAnalyzer(CPastTradesAnalyzer &analyzer, CLearning &learn)//Assigns TradeAnalyzer to object
		{
			m_PastTradesAnalyzer = &analyzer;
			m_PastTradesLearn = &learn;
		}

		void AssignOpenTradesObject(COpenTrades &openTrades)//Asigning objet for storing ids of opened trades
		{
			m_OpenTrades = &openTrades;
		}

		void AssignReverseTradesMethodG(CBool &reverseTradesMethodG)
		{
			m_ReverseTradesAndLotSize = &reverseTradesMethodG;
		}

        bool Init(CCurrencyPairs &pairs, IDateTime &date, CParams &params)//Inits object
        {
            m_WasMinimumMaximumInForBuy.Resize(0);
            m_WasMinimumMaximumInForSell.Resize(0);
            m_WasMinimumMaximumOtherInForBuy.Resize(0);
            m_WasMinimumMaximumOtherInForSell.Resize(0);
            m_AnalyzersForToBuy.Resize(0);
            m_AnalyzersForToSell.Resize(0);
            m_AnalyzersOtherForToBuy.Resize(0);
            m_AnalyzersOtherForToSell.Resize(0);
            m_ValuesAnalyzersForToBuy.Resize(0);
            m_ValuesAnalyzersForToSell.Resize(0);
            m_ValuesAnalyzersOtherForToBuy.Resize(0);
            m_ValuesAnalyzersOtherForToSell.Resize(0);
            
            m_WasMinimumMaximumInForBuy.Resize((*m_Pairs).AmountCurrencyPairsToBuy());
            m_WasMinimumMaximumInForSell.Resize((*m_Pairs).AmountCurrencyPairsToSell());
            m_WasMinimumMaximumOtherInForBuy.Resize((*m_Pairs).AmountOtherSymbolsToBuy());
            m_WasMinimumMaximumOtherInForSell.Resize((*m_Pairs).AmountOtherSymbolsToSell());
            m_AnalyzersForToBuy.Resize((*m_Pairs).AmountCurrencyPairsToBuy());
            m_AnalyzersForToSell.Resize((*m_Pairs).AmountCurrencyPairsToSell());
            m_AnalyzersOtherForToBuy.Resize((*m_Pairs).AmountOtherSymbolsToBuy());
            m_AnalyzersOtherForToSell.Resize((*m_Pairs).AmountOtherSymbolsToSell());
            m_ValuesAnalyzersForToBuy.Resize((*m_Pairs).AmountCurrencyPairsToBuy());
            m_ValuesAnalyzersForToSell.Resize((*m_Pairs).AmountCurrencyPairsToSell());
            m_ValuesAnalyzersOtherForToBuy.Resize((*m_Pairs).AmountOtherSymbolsToBuy());
            m_ValuesAnalyzersOtherForToSell.Resize((*m_Pairs).AmountOtherSymbolsToSell());
                          
            m_Params = &params;

            GetInteroperability().RefreshRates();  
                      
            double val;
            bool was = false;
                        
            for(int i=0;i<pairs.AmountCurrencyPairsToBuy();i++)
            {
				IString pair(pairs.ReturnCurrencyPairToBuyNr(i));

                val = CValues::GetPriceAtTime(pair, date);
                if(val!=0.0)
                {
                    was = true;
                }
                (*(m_AnalyzersForToBuy.GetPointerToValue(i))).SetDefaultMinimumMaximum(val, val * CConfig::PercentageOfChange, date, 0);
            }
            for(int i=0;i<pairs.AmountCurrencyPairsToSell();i++)
            {
				IString pair(pairs.ReturnCurrencyPairToSellNr(i));

                val = CValues::GetPriceAtTime(pair, date);
                if(val!=0.0)
                {
                    was = true;
                }
                (*(m_AnalyzersForToSell.GetPointerToValue(i))).SetDefaultMinimumMaximum(val, val * CConfig::PercentageOfChange, date, 0);
            }
            for(int i=0;i<pairs.AmountOtherSymbolsToBuy();i++)
            {
				IString pair(pairs.ReturnOtherSymbolToBuyNr(i));

                val = CValues::GetPriceAtTime(pair, date);
                if(val!=0.0)
                {
                    was = true;
                }
                (*(m_AnalyzersOtherForToBuy.GetPointerToValue(i))).SetDefaultMinimumMaximum(val, val * CConfig::PercentageOfChange, date, 0);
            }
            for(int i=0;i<pairs.AmountOtherSymbolsToSell();i++)
            {
				IString pair(pairs.ReturnOtherSymbolToSellNr(i));

                val = CValues::GetPriceAtTime(pair, date);
                if(val!=0.0)
                {
                    was = true;
                }
                (*(m_AnalyzersOtherForToSell.GetPointerToValue(i))).SetDefaultMinimumMaximum(val, val * CConfig::PercentageOfChange, date, 0);
            }
            
            m_Pairs = &pairs;
            return was;
        }
    
		void AddToAnalysis(IDateTime &data, bool Symbol, bool Buy)//Adds to analysis one type of Symbols
		{
			double val;
			double diffrence = 0.0;
			double valAfterSpread;
			CPastPoint point;

			for (int i = 0; i < (*m_Pairs).AmountSymbols(Symbol, Buy); i++)
			{
				IString pair((*m_Pairs).ReturnSymbol(i, Symbol, Buy));

				val = CValues::Price(pair, Buy);
				valAfterSpread = CValues::Price(pair, Buy);
				(*WasMinimumMaximum(Symbol, Buy)).SetValue(i, (*((*Analyzers(Symbol, Buy)).GetPointerToValue(i))).CheckForMinimumMaximum(val, data, 0));
				if ((*((*ValuesAnalyzers(Symbol, Buy)).GetPointerToValue(i))).CheckProfit(pair, valAfterSpread, diffrence, true))
				{
					m_SearchingProfit--;
					if (diffrence > 0.0)
					{
						(*m_Params).SuccessfullInRow++;
						(*m_Params).SuccessfullFinds++;
						if ((*m_Params).SuccessfullInRow >= CConfig::SuccededInRow)
						{
							(*m_Params).FailedInRow = 0;
							if (CConfig::UseSingleInRow)
							{
								(*(*m_Pairs).ReturnMultiTo(i, Symbol, Buy)).SetSuccedded(pair, true);
							}
							(*m_Params).StartedTrading = true;
						}
						GetLogger().LogSpecial(IString("Searching Succeded. ") + pair + " Value : " + GetInteroperability().DoubleToString(val));
						GetLogger().LogSpecial(IString("SuccessfullInRow : ") + GetInteroperability().DoubleToString((*m_Params).SuccessfullInRow) + ", FailedInRow : " + GetInteroperability().DoubleToString((*m_Params).FailedInRow));
					}
					else
					{
						(*m_Params).FailedInRow++;
						(*m_Params).FailedFinds++;
						if ((*m_Params).FailedInRow >= CConfig::FailedInRow)
						{
							(*m_Params).SuccessfullInRow = 0;
							if (CConfig::UseSingleInRow)
							{
								(*(*m_Pairs).ReturnMultiTo(i, Symbol, Buy)).SetSuccedded(pair, false);
							}
							(*m_Params).StartedTrading = true;
						}
						GetLogger().LogSpecial(IString("Searching Failed. ") + pair);
						GetLogger().LogSpecial(IString("SuccessfullInRow : ") + GetInteroperability().DoubleToString((*m_Params).SuccessfullInRow) + ", FailedInRow : " + GetInteroperability().DoubleToString((*m_Params).FailedInRow));
					}
				}
				if ((*WasMinimumMaximum(Symbol, Buy)).GetValue(i) != CAnalyzer::NO_RESULT)
				{
					if ((*m_Pairs).TradablePeriod() && (*m_Pairs).IsSymbolPredictedProfitable(i, Symbol, Buy))
					{
						double diff = GetInteroperability().MathAbs(val - (*WasMinimumMaximum(Symbol, Buy)).GetValue(i));
						IDateTime mmDate((*((*Analyzers(Symbol, Buy)).GetPointerToValue(i))).ReturnLastMinimumMaximumDate());

						if (ProceedWithOperation(Buy, pair, mmDate, data, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy)))
						{
							if (CConfig::UseSellBuyNeuralNetworks)
							{
								GetLogger().LogSpecial(IString("Network allowed for : ") + pair);
							}
							if (ProceedWithTrade(pair, i, Symbol, Buy, mmDate, data, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy)))
							{
								if (CConfig::UseIndividualNeuralNetworks)
								{
									GetLogger().LogSpecial(IString("Symbol Network allowed for : ") + pair);
								}
								if (CTrading::ProceedWithSearching((*m_Pairs), !Symbol, i, Buy, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy)))
								{
									if (m_SearchingProfit < CConfig::SingleTradesAmount)
									{
										bool trendReverseTrades;

										trendReverseTrades = CheckTrendReverseTrades();
										AfterTrendReverseTrades();
										(*((*ValuesAnalyzers(Symbol, Buy)).GetPointerToValue(i))).SearchForProfit(pair, val, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy), Buy, point, false, trendReverseTrades, (*m_Params).AfterTrendReverseTrades, (*m_Params).AutomaticParamsReverseTrades);
										GetLogger().LogSpecial(IString("SearchForProfit upprLimit : ") + GetInteroperability().DoubleToString((*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy)) + ", lowerLimit : " + GetInteroperability().DoubleToString((*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy)));
										GetLogger().LogSpecial(IString("Started Searching. ") + pair + ", Diff : " + GetInteroperability().DoubleToString(diff));
										m_SearchingProfit++;

										bool methodCmethodDReverseTrades = false;

										if (trendReverseTrades && CConfig::UseTrendFromLogs)
										{
											methodCmethodDReverseTrades = !methodCmethodDReverseTrades;
										}
										if((*m_Params).AfterTrendReverseTrades && CConfig::UseAfterTrend)
										{
											methodCmethodDReverseTrades = !methodCmethodDReverseTrades;
										}
										if((*m_Params).AutomaticParamsReverseTrades && CConfig::UseAfterTrend && CConfig::UseAfterTrendAutomaticParams)
										{
											methodCmethodDReverseTrades = !methodCmethodDReverseTrades;
										}
										
										if (CConfig::ByPassInRowCheck || ((*m_Params).StartedTrading && (((!ReverseTrades((*(*m_Pairs).ReturnMultiTo(i, Symbol, Buy)).ReverseTrades)) && (*m_Params).SuccessfullInRow >= CConfig::SuccededInRow && (*m_Params).FailedInRow <= CConfig::FailedInRow) || ((ReverseTrades((*(*m_Pairs).ReturnMultiTo(i, Symbol, Buy)).ReverseTrades)) && (*m_Params).FailedInRow >= CConfig::SuccededInRow && (*m_Params).SuccessfullInRow <= CConfig::FailedInRow))))
										{
											if (CheckParamsVector())
											{
												if (!CConfig::UseRules || (CConfig::UseRules && IsRuleHit(methodCmethodDReverseTrades)))
												{
													bool methodFReverseTrades = false;
													
													CheckDistance(Symbol, Buy, i, methodFReverseTrades);
													if (!CConfig::UseCalculateDistance || (CConfig::UseCalculateDistance && (methodCmethodDReverseTrades == methodFReverseTrades || (!CConfig::UseTrendFromLogs && !CConfig::UseAfterTrend))))
													{
														int tradeNr = CTrading::SendOrder(*m_Pairs, !Symbol, i, Buy, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy), trendReverseTrades, (*m_Params).AfterTrendReverseTrades, (*m_Params).AutomaticParamsReverseTrades, methodFReverseTrades, (*m_ReverseTradesAndLotSize).m_Bool, (*m_ReverseTradesAndLotSize).m_Double);
														if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_STOP || CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_PROFIT)
														{
															(*m_OpenTrades).AddOpenTrade(tradeNr);
				   										    tradeNr = CTrading::SendOrder(*m_Pairs, !Symbol, i, !Buy, (*m_Pairs).ReturnUpperLimit(val, diff, i, Symbol, Buy), (*m_Pairs).ReturnLowerLimit(val, diff, i, Symbol, Buy), trendReverseTrades, (*m_Params).AfterTrendReverseTrades, (*m_Params).AutomaticParamsReverseTrades, methodFReverseTrades, (*m_ReverseTradesAndLotSize).m_Bool, (*m_ReverseTradesAndLotSize).m_Double);
															(*m_OpenTrades).AddReverseOpenTrade(tradeNr);
														}
													}
												}
											}
											else
											{
												GetLogger().LogSpecial(IString("Params Vector not filled, bypassing Trade : ") + pair);
											}
										}
										else
										{
											GetLogger().LogSpecial(IString("Order Filtered Out. ") + pair);
										}
									}
									else
									{
										GetLogger().LogSpecial(IString("Searching Filtered Out. ") + pair);
									}
								}
								else
								{
									GetLogger().LogSpecial(IString("Ignoring Searching for : ") + pair);
								}
							}
							else
							{
								if (CConfig::UseIndividualNeuralNetworks)
								{
									GetLogger().LogSpecial(IString("Symbol Network disallowed for : ") + pair);
								}
							}
						}
						else
						{
							if (CConfig::UseSellBuyNeuralNetworks)
							{
								GetLogger().LogSpecial(IString("Network disallowed for : ") + pair);
							}
						}
					}
				}
			}
		}

        void AddToAnalysis(IDateTime &data)//Adds to analysis all Symbols
        {
            SetWasMMtoFalse();
			AddToAnalysis(data, true, true);
			AddToAnalysis(data, true, false);
			AddToAnalysis(data, false, true);
			AddToAnalysis(data, false, false);
        }   
};
//+------------------------------------------------------------------+
#endif