/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

#ifndef FILE_VALUE_ANALYZER
#define FILE_VALUE_ANALYZER

#include "Analyzer.h"
#include "Values.h"
#include "PastPoint.h"
#include "../../Interoperability/String.h"
#include "Params.h"
#include "../LibrariesInteroperability.h"
#include "../LibrariesLogger.h"

//+------------------------------------------------------------------+
class CValueAnalyzer {
private:
	long static ProfitSearchNr;

	double m_LastEstablishingMinimumMaximum;
	bool m_SearchingForProfit;
	double m_UpperLimit;
	double m_LowerLimit;

	bool m_Buy;
	bool m_BuyRT;

	double m_Gain;
	double m_Loss;

	int m_GainCount;
	int m_LossCount;

	bool m_DirectionAchivedGuard;
	int m_DirectionAchivedAmount;

	double m_ValueFirstToReach;
	bool m_ValueFirstReached;
	
	double m_ValueSecondToReach;
	bool m_ValueSecondReached;
	
	double m_ValueThirdToReach;
	bool m_ValueThirdReached;
	
	double m_ValueFourthToReach;
	bool m_ValueFourthReached;

	long m_ProfitSearchNr;

	static CParams *m_Params;

	bool m_Initial;

	void Reset()
	{
		m_Gain = 0;
		m_Loss = 0;
		m_GainCount = 0;
		m_LossCount = 0;
		m_DirectionAchivedAmount = 0;
	}

	CPastPoint m_Point;

	double ValueToReach(bool buy, double percent, double lastEstablishingMinimumMaximum, double upperLimit, double lowerLimit)
	{
		if (buy)
		{
			return lastEstablishingMinimumMaximum + (upperLimit - lastEstablishingMinimumMaximum) * percent;
		}
		else
		{
			return lastEstablishingMinimumMaximum - (lastEstablishingMinimumMaximum - lowerLimit) * percent;
		}
	}

	void ValueReached(bool buy, double val, double valueToReach, bool &valueReached, IString &msg)
	{
		if (buy)
		{
			if (!valueReached)
			{
				if (val >= valueToReach)
				{
					GetLogger().LogSpecial(msg);
					valueReached = true;
				}
			}
		}
		else
		{
			if (!valueReached)
			{
				if (val <= valueToReach)
				{
					GetLogger().LogSpecial(msg);
					valueReached = true;
				}
			}
		}
	}

public:
	void static SetParams(CParams &params)
	{
		m_Params = &params;
	}

	CValueAnalyzer()
	{
		Reset();
	}

	bool CheckProfit(IString &pair, double val, double &diff, bool log)
	{
		if (m_SearchingForProfit)
		{
			IString msgFirst("First Percent Value Min Reached."); msgFirst += IString(" Profit Search Nr : "); msgFirst += GetInteroperability().IntegerToString(m_ProfitSearchNr);
		   IString msgSecond("Second Percent Value Min Reached."); msgSecond += IString(" Profit Search Nr : "); msgSecond += GetInteroperability().IntegerToString(m_ProfitSearchNr);
		   IString msgThird("Third Percent Value Min Reached."); msgThird += IString(" Profit Search Nr : "); msgThird += GetInteroperability().IntegerToString(m_ProfitSearchNr);
		   IString msgFourth("Fourth Percent Value Min Reached."); msgFourth += IString(" Profit Search Nr : "); msgFourth += GetInteroperability().IntegerToString(m_ProfitSearchNr);
		   
			ValueReached(m_BuyRT, val, m_ValueFirstToReach, m_ValueFirstReached, msgFirst);
			ValueReached(m_BuyRT, val, m_ValueSecondToReach, m_ValueSecondReached, msgSecond);
			ValueReached(m_BuyRT, val, m_ValueThirdToReach, m_ValueThirdReached, msgThird);
			ValueReached(m_BuyRT, val, m_ValueFourthToReach, m_ValueFourthReached, msgFourth);

			if (m_Buy)
			{
				if (!m_DirectionAchivedGuard)
				{
					if (val > m_LastEstablishingMinimumMaximum)
					{
						m_DirectionAchivedGuard = true;
						m_DirectionAchivedAmount++;
					}
				}

				diff = val - m_LastEstablishingMinimumMaximum;

				if (val >= m_UpperLimit)
				{
					m_Gain += (diff);
					m_GainCount++;
					m_SearchingForProfit = false;
					m_Point.Successful = true;
					if (log)
					{
						GetLogger().LogSpecial(IString("Profit found, Symbol : ") + pair + ", Operation : " + (m_Buy ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						if (m_BuyRT)
						{
							GetLogger().LogSpecial(IString("Profit_found_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
						else
						{
							GetLogger().LogSpecial(IString("Profit_lost_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
					}
					if (!m_Initial)
					{
						(*m_Params).AddPastTrade(true);
						(*m_Params).AddPastTradeWithRT(m_BuyRT);
					}
					return true;
				}
				else if (val < m_LowerLimit)
				{
					m_Loss += (GetInteroperability().MathAbs(diff));
					m_LossCount++;
					m_SearchingForProfit = false;
					m_Point.Successful = false;
					if (log)
					{
						GetLogger().LogSpecial(IString("Profit lost, Symbol : ") + pair + ", Operation : " + (m_Buy ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						if (m_BuyRT)
						{
							GetLogger().LogSpecial(IString("Profit_lost_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
						else
						{
							GetLogger().LogSpecial(IString("Profit_found_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
					}
					if (!m_Initial)
					{
						(*m_Params).AddPastTrade(false);
						(*m_Params).AddPastTradeWithRT(!m_BuyRT);
					}
					return true;
				}
			}
			else
			{
				if (!m_DirectionAchivedGuard)
				{
					if (val < m_LastEstablishingMinimumMaximum)
					{
						m_DirectionAchivedGuard = true;
						m_DirectionAchivedAmount++;
					}
				}

				diff = m_LastEstablishingMinimumMaximum - val;

				if (val < m_LowerLimit)
				{
					m_Gain += (diff);
					m_GainCount++;
					m_SearchingForProfit = false;
					m_Point.Successful = true;
					if (log)
					{
						GetLogger().LogSpecial(IString("Profit found, Symbol : ") + pair + ", Operation : " + (m_Buy ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", TP : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						if (!m_BuyRT)
						{
							GetLogger().LogSpecial(IString("Profit_found_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
						else
						{
							GetLogger().LogSpecial(IString("Profit_lost_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
					}
					if (!m_Initial)
					{
						(*m_Params).AddPastTrade(true);
						(*m_Params).AddPastTradeWithRT(!m_BuyRT);
					}
					return true;
				}
				else if (val >= m_UpperLimit)
				{
					m_Loss += (GetInteroperability().MathAbs(diff));
					m_LossCount++;
					m_SearchingForProfit = false;
					m_Point.Successful = false;
					if (log)
					{
						GetLogger().LogSpecial(IString("Profit lost, Symbol : ") + pair + ", Operation : " + (m_Buy ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", TP : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						if (!m_BuyRT)
						{
							GetLogger().LogSpecial(IString("Profit_lost_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
						else
						{
							GetLogger().LogSpecial(IString("Profit_found_RT, Symbol : ") + pair + ", Operation : " + (m_BuyRT ? "Buy" : "Sell") + ", SL : " + GetInteroperability().DoubleToString(m_UpperLimit) + ", TP : " + GetInteroperability().DoubleToString(m_LowerLimit) + ", Profit Search End Nr : " + GetInteroperability().IntegerToString(m_ProfitSearchNr));
						}
					}
					if (!m_Initial)
					{
						(*m_Params).AddPastTrade(false);
						(*m_Params).AddPastTradeWithRT(m_BuyRT);
					}
					return true;
				}
			}
		}
		return false;
	}

	void Log(IString &symbol)
	{
		GetLogger().Log(symbol + " Gain : " + GetInteroperability().DoubleToString(m_Gain) + " Loss : " + GetInteroperability().DoubleToString(m_Loss) + " Profit : " + GetInteroperability().DoubleToString(Profit()) + " Gain Count : " + GetInteroperability().IntegerToString(m_GainCount) + " Loss Count : " + GetInteroperability().IntegerToString(m_LossCount));
	}

	int GainCount()
	{
		return m_GainCount;
	}

	int LossCount()
	{
		return m_LossCount;
	}

	double Profit()
	{
		return m_Gain - m_Loss;
	}

	double ProfitMinusSpread(IString &symbol)
	{
		double spread = CValues::Spread(symbol);
		double spreads = spread * (m_GainCount + m_LossCount);

		return Profit() - spreads;
	}

	double ProfitGains()
	{
		double count;

		if (m_LossCount > 0)
		{
			count = m_GainCount / (double)m_LossCount;
		}
		else
		{
			count = m_GainCount;
		}
		return count * Profit();
	}

	bool IsSearchingForProfit()
	{
		return m_SearchingForProfit;
	}

	CPastPoint ReturnPoint()
	{
		return m_Point;
	}

	void SearchForProfit(IString &pair, double lastEstablishingMinimumMaximum, double upperLimit, double lowerLimit, bool buy, CPastPoint &point, bool initial, bool reverseTrend, bool afterTrendReverse, bool automaticReverse)	
	{
		m_UpperLimit = upperLimit;
		m_LowerLimit = lowerLimit;
		m_SearchingForProfit = true;
		m_LastEstablishingMinimumMaximum = lastEstablishingMinimumMaximum;
		m_Buy = buy;
		m_BuyRT = buy;
		m_Point = point;
		m_DirectionAchivedGuard = false;
		m_ProfitSearchNr = ProfitSearchNr;
		ProfitSearchNr++;

		if (reverseTrend)
		{
			m_BuyRT = !m_BuyRT;
		}
			   
		if (initial)
		{
			m_ValueFirstReached = true;
			m_ValueSecondReached = true;
			m_ValueThirdReached = true;
			m_ValueFourthReached = true;
		}
		else
		{
			GetLogger().LogSpecial(IString("Profit Search Start Nr : ") + GetInteroperability().IntegerToString(m_ProfitSearchNr));

			m_ValueFirstReached = false;
			m_ValueSecondReached = false;
			m_ValueThirdReached = false;
			m_ValueFourthReached = false;

			bool buyRTAll = m_BuyRT;

			if (afterTrendReverse)
			{
				buyRTAll = !buyRTAll;
			}

			if (automaticReverse)
			{
				buyRTAll = !buyRTAll;
			}

			GetLogger().LogSpecial(IString("Place Order, Symbol : ") + pair + IString(", Buy Price : ") + GetInteroperability().DoubleToString(CValues::BuyPrice(pair)) + IString(", Sell Price : ") + GetInteroperability().DoubleToString(CValues::SellPrice(pair)) + IString(", Type : ") + (m_Buy ? IString("Buy") : IString("Sell")) + IString(", Type_RT : ") + (m_BuyRT ? IString("Buy") : IString("Sell")) + IString(", Type_RT_All : ") + (buyRTAll ? IString("Buy") : IString("Sell")) + IString(", Upper Limit : ") + GetInteroperability().DoubleToString(upperLimit) + IString(", Lower Limit : ") + GetInteroperability().DoubleToString(lowerLimit));
		}

		m_Initial = initial;

		m_ValueFirstToReach = ValueToReach(m_BuyRT, CConfig::MinFirstPercentOfProfitReached, lastEstablishingMinimumMaximum, upperLimit, lowerLimit);
		m_ValueSecondToReach = ValueToReach(m_BuyRT, CConfig::MinSecondPercentOfProfitReached, lastEstablishingMinimumMaximum, upperLimit, lowerLimit);
		m_ValueThirdToReach = ValueToReach(m_BuyRT, CConfig::MinThirdPercentOfProfitReached, lastEstablishingMinimumMaximum, upperLimit, lowerLimit);
		m_ValueFourthToReach = ValueToReach(m_BuyRT, CConfig::MinFourthPercentOfProfitReached, lastEstablishingMinimumMaximum, upperLimit, lowerLimit);
	}

	double DirectionAchivedPercent()
	{
		return ((double)(m_DirectionAchivedAmount)) / ((double)(m_GainCount + m_LossCount));
	}
};
long CValueAnalyzer::ProfitSearchNr = 1;
CParams* CValueAnalyzer::m_Params = NULL;
//+------------------------------------------------------------------+
#endif