/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

#ifndef FILE_CALCULUS_ROBOT_EXPERT
#define FILE_CALCULUS_ROBOT_EXPERT

#include "CR Includes\\Config.h"
#include "CR Includes\\Logger.h"
#include "CR Includes\\Logs.h"
#include "CR Includes\\PastTradesAnalyzer.h"
#include "CR Includes\\CurrencyPairs.h"
#include "CR Includes\\Timer.h"
#include "CR Includes\\Trading.h"
#include "CR Includes\\MultiAnalyzerInit.h"
#include "CR Includes\\MultiAnalyzerProcess.h"
#include "CR Includes\\MultiAnalyzerPast.h"
#include "CR Includes\\Learning.h"
#include "CR Includes\\LinesProcessor.h"
#include "CR Includes\\Fcnnhn\\Vectors.h"
#include "CR Includes\\Params.h"
#include "CR Includes\\FarPastDataFile.h"
#include "../Interoperability/Interoperability.h"
#include "../Interoperability/GenericObjectArray.h"
#include "../Interoperability/DateTime.h"
#include "CR Includes/DistancesProvider.h"
#include "CR Includes\\PastProfits.h"
#include "LibrariesInteroperability.h"
#include "LibrariesLogger.h"
#include "CR Includes\\OpenTrades.h"
#include "CR Includes\\Bool.h"

//+------------------------------------------------------------------+
class CCalculusRobotExpert{
    private:
        static const int MaxLeverage;
        IDateTime m_LastInitialization;
		IDateTime m_LastFarPastDate;
		IDateTime m_LastPastProfit;
		CCurrencyPairs *m_Pairs;
        CTimer m_Timer;
        CValues m_Values;
        CMultiAnalyzerInit m_AnlyzerInit;
        CMultiAnalyzerProcess *m_AnalyzerProcess;
		CMultiAnalyzerPast m_AnalyzerPast;
        CLearning m_NetBuys;
        CLearning m_NetSells;
        CLearning m_NetTestBuys;
        CLearning m_NetTestSells;
        CParams *m_Params;
		CPastTradesAnalyzer m_PastTradesAnalyzer;
		CLearning m_PastTradesLearn;
		CDistancesProvider m_DistancesProvider;
		CValues m_PastValues;
		double m_MaxProfitValue;
		double m_MinProfitValue;
		int m_TrailSpeedLogCountDown;
		CBool m_ReverseTradesAndLotSize;
		double m_BalanceLastMax;
        
        bool IsValid()
        {
            if(CConfig::UseSellBuyNeuralNetworks)
            {
                return m_NetBuys.IsValid() || m_NetSells.IsValid();
            }
            else
            {
                return true;
            }
        }
        
        bool LearnNetwork(CLearning &learning, IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors, double testPercent)
        {
            learning.Init((*(lines.GetPointerToValue(0))).Points.Size() + 1 + 1, 1, CConfig::HiddenLayersAmount, CConfig::ActivationFunctionDimension);
            learning.Learn(vectors, testPercent);
            return learning.IsValid();
        }
        
        bool LearnSymbolNetwork(CLearning &learning, IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors, double testPercent)
        {
            learning.Init((*(lines.GetPointerToValue(0))).Points.Size() + 1 + 1, 1, CConfig::HiddenLayersAmountForSymbolNetwork, CConfig::ActivationFunctionDimensionForSymbolNetwork);
            learning.Learn(vectors, testPercent);
            return learning.IsValid();
        }
        
        bool LearnTestNetForBuy(IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors)
        {
            GetLogger().LogSpecial(IString("Learning Test Network to Buy."));
            for(int i=0;i<CConfig::NeuralNetworkLearnAttempts;i++)
            {
                GetLogger().LogSpecial(GetInteroperability().IntegerToString(i+1) + "|" + GetInteroperability().IntegerToString(CConfig::NeuralNetworkLearnAttempts) + " Learn Attempt.");
                if(LearnNetwork(m_NetTestBuys, lines, vectors, CConfig::TestVectorsPercent))
                {
                    return true;
                }
            }
            return false;
        }

        bool LearnTestNetForSell(IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors)
        {
            GetLogger().LogSpecial(IString("Learning Test Network to Sell."));
            for(int i=0;i<CConfig::NeuralNetworkLearnAttempts;i++)
            {
                GetLogger().LogSpecial(GetInteroperability().IntegerToString(i+1) + "|" +GetInteroperability(). IntegerToString(CConfig::NeuralNetworkLearnAttempts) + " Learn Attempt.");
                if(LearnNetwork(m_NetTestSells, lines, vectors, CConfig::TestVectorsPercent))
                {
                    return true;
                }
            }
            return false;
        }

        bool LearnNetForBuy(IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors)
        {
            GetLogger().LogSpecial(IString("Learning Network to Buy."));
            for(int i=0;i<CConfig::NeuralNetworkLearnAttempts;i++)
            {
                GetLogger().LogSpecial(GetInteroperability().IntegerToString(i+1) + "|" + GetInteroperability().IntegerToString(CConfig::NeuralNetworkLearnAttempts) + " Learn Attempt.");
                if(LearnNetwork(m_NetBuys, lines, vectors, 0.0))
                {
                    return true;
                }
            }
            return false;
        }

        bool LearnNetForSell(IGenericObjectArray<CLineProcessed> &lines, CVectors &vectors)
        {
            GetLogger().LogSpecial(IString("Learning Network to Sell."));
            for(int i=0;i<CConfig::NeuralNetworkLearnAttempts;i++)
            {
                GetLogger().LogSpecial(GetInteroperability().IntegerToString(i+1) + "|" + GetInteroperability().IntegerToString(CConfig::NeuralNetworkLearnAttempts) + " Learn Attempt.");
                if(LearnNetwork(m_NetSells, lines, vectors, 0.0))
                {
                    return true;
                }
            }
            return false;
        }
        
        void ChoosePastPoints(int symbolNr, bool Symbol, bool SymbolBuy, IGenericObjectArray<CPastPoint> &points)
        {
            int am = 0;
            int pos = 0;
        
            if(SymbolBuy)
            {
                for(int i=0;i<m_AnlyzerInit.PastPointsBuys.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).Symbol == Symbol && (*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).SymbolNr == symbolNr)
                    {
                        am++;
                    }                        
                }
                points.Resize(am);
                for(int i=0;i<m_AnlyzerInit.PastPointsBuys.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).Symbol == Symbol && (*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).SymbolNr == symbolNr)
                    {
                        (*(points.GetPointerToValue(pos))) = (*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i)));
                        pos++;
                    }                        
                }
            }
            else
            {
                for(int i=0;i<m_AnlyzerInit.PastPointsSells.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).Symbol == Symbol && (*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).SymbolNr == symbolNr)
                    {
                        am++;
                    }                        
                }
                points.Resize(am);
                for(int i=0;i<m_AnlyzerInit.PastPointsSells.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).Symbol == Symbol && (*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).SymbolNr == symbolNr)
                    {
                        (*(points.GetPointerToValue(pos))) = (*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i)));
                        pos++;
                    }                        
                }                    
            }
        }
        
        bool LearnSymbolNet(CLearning &learn, IGenericObjectArray<CPastPoint> &points)
        {
            IGenericObjectArray<CLineProcessed> differentialLines;
            CLinesProcessor::GenerateDifferentialLines(points, differentialLines, m_Values);
        
            CVectors vectors;
            
            int pos = 0;
            double val, diff;
            int SymbolNr = 0;
                    
            if(differentialLines.Size() > 0)
            {
                vectors.Init(differentialLines.Size(), (*(differentialLines.GetPointerToValue(0))).Points.Size() + 1 + 1, 1);
                           
                for(int i=0;i<points.Size();i++)
                {
                    if((*(points.GetPointerToValue(i))).IsValid())
                    {     
                        SymbolNr = (*(points.GetPointerToValue(i))).SymbolNr;
                        for(int j=0;j<(*(differentialLines.GetPointerToValue(0))).Points.Size();j++)
                        {
                            (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue(j, (*(differentialLines.GetPointerToValue(pos))).Points.GetValue(j));
                        }
                        if((*(points.GetPointerToValue(i))).Successful)
                        {
                            (*(vectors.Vectors.GetPointerToValue(pos))).Output.SetValue(0, 1.0);
                        }
                        else
                        {                    
                            (*(vectors.Vectors.GetPointerToValue(pos))).Output.SetValue(0, -1.0);
                        }
                        if((*(points.GetPointerToValue(i))).SymbolBuy)
                        {
                            if((*(points.GetPointerToValue(i))).Symbol)
                            {
                                val = m_Values.ReturnPastValueNrToBuy((*(points.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                                diff = GetInteroperability().MathAbs(val - m_Values.ReturnPastValueNrToBuy((*(points.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, true, true));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, true, true));
                            }
                            else
                            {
                                val = m_Values.ReturnOtherPastValueNrToBuy((*(points.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                                diff = GetInteroperability().MathAbs(val - m_Values.ReturnOtherPastValueNrToBuy((*(points.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, false, true));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, false, true));
                            }
                        }
                        else
                        {
                            if((*(points.GetPointerToValue(i))).Symbol)
                            {
                                val = m_Values.ReturnPastValueNrToSell((*(points.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                                diff = GetInteroperability().MathAbs(val - m_Values.ReturnPastValueNrToSell((*(points.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, true, false));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, true, false));
                            }
                            else
                            {
                                val = m_Values.ReturnOtherPastValueNrToSell((*(points.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                                diff = GetInteroperability().MathAbs(val - m_Values.ReturnOtherPastValueNrToSell((*(points.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, false, false));
                                (*(vectors.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLines.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, false, false));
                            }
                        }                        
                        pos++;
                    }
                }
                
                IString name;
                
                if((*(points.GetPointerToValue(0))).Symbol)
                {
                    if((*(points.GetPointerToValue(0))).SymbolBuy)
                    {
                        name.AssignString((*m_Pairs).ReturnCurrencyPairToBuyNr(SymbolNr));
                    }
                    else
                    {
                        name.AssignString((*m_Pairs).ReturnCurrencyPairToSellNr(SymbolNr));
                    }
                }
                else
                {
                    if((*(points.GetPointerToValue(0))).SymbolBuy)
                    {
                        name.AssignString((*m_Pairs).ReturnOtherSymbolToBuyNr(SymbolNr));
                    }
                    else
                    {
                        name.AssignString((*m_Pairs).ReturnOtherSymbolToSellNr(SymbolNr));
                    }
                }
                
                GetLogger().LogSpecial(IString("Learning Symbol Network for ") + name);
                if(LearnSymbolNetwork(*(*m_Pairs).ReturnLearningObject((*(points.GetPointerToValue(0))).SymbolNr, (*(points.GetPointerToValue(0))).Symbol, (*(points.GetPointerToValue(0))).SymbolBuy), differentialLines, vectors, 0.0))
                {
                    return true;
                }
            }
            return false;
        }
        
        void LearnSymbolNets()
        {
            IGenericObjectArray<CPastPoint> points;
            int learned = 0;
            int all = 0;
        
            GetLogger().LogSpecial(IString("Learning Symbol Nets."));
            for(int i=0;i<(*m_Pairs).AmountCurrencyPairsToBuy();i++)
            {
                if((*(*m_Pairs).ReturnMultiTo(i, true, true)).IsTradable)
                {
                    ChoosePastPoints(i, true, true, points);
                    if(LearnSymbolNet(*((*m_Pairs).ReturnLearningObject(i, true, true)), points))
                    {
                        learned++;
                    }
                    all++;
                }
            }
            for(int i=0;i<(*m_Pairs).AmountCurrencyPairsToSell();i++)
            {
                if((*(*m_Pairs).ReturnMultiTo(i, true, false)).IsTradable)
                {
                    ChoosePastPoints(i, true, false, points);
                    if(LearnSymbolNet(*((*m_Pairs).ReturnLearningObject(i, true, false)), points))
                    {
                        learned++;
                    }
                    all++;
                }
            }
            for(int i=0;i<(*m_Pairs).AmountOtherSymbolsToBuy();i++)
            {
                if((*(*m_Pairs).ReturnMultiTo(i, false, true)).IsTradable)
                {
                    ChoosePastPoints(i, false, true, points);
                    if(LearnSymbolNet(*((*m_Pairs).ReturnLearningObject(i, false, true)), points))
                    {
                        learned++;
                    }
                    all++;
                }                
            }
            for(int i=0;i<(*m_Pairs).AmountOtherSymbolsToSell();i++)
            {
                if((*(*m_Pairs).ReturnMultiTo(i, false, false)).IsTradable)
                {
                    ChoosePastPoints(i, false, false, points);
                    if(LearnSymbolNet(*((*m_Pairs).ReturnLearningObject(i, false, false)), points))
                    { 
                        learned++;
                    }
                    all++;
                }
            }
            GetLogger().LogSpecial(IString("Symbol Nets Learned : ") + GetInteroperability().IntegerToString(learned) + "|" + GetInteroperability().IntegerToString(all));
        }
        
		void DetermineMethodDPArams()
		{
			if (CConfig::UseAfterTrendAutomaticParams)
			{
				CLogs logs;

				logs.ReadLogs();
				logs.ExtractTrades((*m_Pairs).TradableAmount(), true);
				logs.AssignOrder();

				GetLogger().LogSpecial(IString("Amount of Trades Extracted from logs : ") + GetInteroperability().IntegerToString(logs.TradesAmount()));
				if (logs.TradesAmount() > CConfig::TradesAmountMin)
				{
					GetLogger().LogSpecial(IString("Determining Method D Params."));
				}
				else
				{
					GetLogger().LogSpecial(IString("Determining Method D Params skipped. Extracted Trades Amount is less then : ") + GetInteroperability().IntegerToString(CConfig::TradesAmountMin));
					return;
				}

				IGenericObjectArray<CTrade> trades;

				logs.FillTrades(trades);
				(*m_Params).DetermineMethodDParams(trades);
			}
		}

		void AnalyzePastData()
		{
         	{
				IGenericObjectArray<CPastValues> vals;

				CFarPastDataFile::ReadDataColumnsValues(*m_Pairs, vals);
				if (CFarPastDataFile::IsPastDataLongEnough(vals))
				{
					m_PastValues.AssignPastValues(vals, *m_Pairs);
					GetLogger().LogSpecial(IString("Past Data long enough."));
				}
				else
				{
					GetLogger().LogSpecial(IString("Past Data not long enough."));
					return;
				}
			}

			m_AnalyzerPast.SetValues(m_PastValues);
			m_AnalyzerPast.PreInit(*m_Pairs);
			m_AnalyzerPast.Analyze();

			m_DistancesProvider.AssignPoints(m_AnalyzerPast.PastPointsBuys, m_AnalyzerPast.PastPointsSells, *m_Pairs, m_PastValues);			
			m_DistancesProvider.SetValidity(true);
		}

		void CheckForClosePart(IGenericArray<double> &profits, IGenericArray<int> &numbers, IDateTime &date)
		{
			double profit = 0.0;
			double reverseProfit = 0.0;

            (*((*m_Params).ReturnOpenTrades())).RemoveNotPresent(numbers);

			for (int i = 0; i < numbers.Size(); i++)
			{
				if ((*((*m_Params).ReturnOpenTrades())).IsReverseTrade(numbers.GetValue(i)))
				{
					reverseProfit += profits.GetValue(i);
				}
				else
				{
					profit += profits.GetValue(i);
				}
			}

            (*((*m_Params).ReturnPastProfits())).AddProfit(profit, date, (*CTrading::GetVaribles()).MultivariateTrailSpeedHours());
			if ((*((*m_Params).ReturnPastProfits())).IsValid((*CTrading::GetVaribles()).MultivariateTrailSpeedHours()))
			{
				if ((*CTrading::GetVaribles()).MultivariateTrailSpeed() > (profit - (*((*m_Params).ReturnPastProfits())).PastProfit()) / (*CTrading::GetVaribles()).MultivariateTrailSpeedHours())
				{
					if (profit != 0.0)
					{
						GetLogger().LogSpecial(IString("Closing Opened."));
						GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnOpenTrades());
                        (*((*m_Params).ReturnPastProfits())).EmptyProfits();
					}
				}
			}

            (*((*m_Params).ReturnPastReverseProfits())).AddProfit(reverseProfit, date, (*CTrading::GetVaribles()).MultivariateTrailSpeedHours());
			if ((*((*m_Params).ReturnPastReverseProfits())).IsValid((*CTrading::GetVaribles()).MultivariateTrailSpeedHours()))
			{
				if ((*CTrading::GetVaribles()).MultivariateReverseTrailSpeed() > (reverseProfit - (*((*m_Params).ReturnPastReverseProfits())).PastProfit()) / (*CTrading::GetVaribles()).MultivariateTrailSpeedHours())
				{
					if (reverseProfit != 0.0)
					{
						GetLogger().LogSpecial(IString("Closing Reverse Opened."));
						GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnReverseOpenTrades());
                        (*((*m_Params).ReturnPastReverseProfits())).EmptyProfits();
					}
				}
			}

			if (LogSpeeds())
			{
				if ((*((*m_Params).ReturnPastProfits())).IsValid((*CTrading::GetVaribles()).MultivariateTrailSpeedHours()))
				{
					double speed = (profit - (*((*m_Params).ReturnPastProfits())).PastProfit()) / (*CTrading::GetVaribles()).MultivariateTrailSpeedHours();

					GetLogger().LogSpeeds(IString("Speed : ") + GetInteroperability().DoubleToString(speed));
				}
				if ((*((*m_Params).ReturnPastReverseProfits())).IsValid((*CTrading::GetVaribles()).MultivariateTrailSpeedHours()))
				{
					double reverseSpeed = (reverseProfit - (*((*m_Params).ReturnPastReverseProfits())).PastProfit()) / (*CTrading::GetVaribles()).MultivariateTrailSpeedHours();

					GetLogger().LogSpeeds(IString("Reverse Speed : ") + GetInteroperability().DoubleToString(reverseSpeed));
				}
			}
		}

		void CheckForReverse(double prof)
		{
			if (prof < (*CTrading::GetVaribles()).MultivariateReverseTradesMin())
			{
				GetInteroperability().CloseAll();
				GetLogger().LogSpecial(IString("Closing All Orders and Reversing, Profit : ") + GetInteroperability().DoubleToString(prof));
				m_ReverseTradesAndLotSize.m_Bool = !m_ReverseTradesAndLotSize.m_Bool;
			}
		}

		void CheckForReverse()
		{
			double balance = GetInteroperability().AccountBalance();

			if (balance < m_BalanceLastMax - (*CTrading::GetVaribles()).MultivariateBalanceReverseTradesMin())
			{
				GetInteroperability().CloseAll();
				GetLogger().LogSpecial(IString("Closing All Orders and Reversing, Balance : ") + GetInteroperability().DoubleToString(balance));
				m_ReverseTradesAndLotSize.m_Bool = !m_ReverseTradesAndLotSize.m_Bool;
				m_BalanceLastMax = balance;
			}
			else if (balance > m_BalanceLastMax)
			{
				m_BalanceLastMax = balance;
			}
		}

        void CheckForReverseBalanceSpeed(IDateTime &date)
        {
            double balance = GetInteroperability().AccountBalance();

            (*((*m_Params).ReturnPastBalances())).AddProfit(balance, date, (*CTrading::GetVaribles()).MultivariateSurfingReverseBalanceSpeedHours());
            if ((*((*m_Params).ReturnPastBalances())).IsSpeedValid((*CTrading::GetVaribles()).MultivariateSurfingReverseBalanceSpeedHours()))
            {
                double speed = (*((*m_Params).ReturnPastBalances())).Speed((*CTrading::GetVaribles()).MultivariateSurfingReverseBalanceSpeedHours());

                if (speed < (*CTrading::GetVaribles()).MultivariateSurfingReverseBalanceSpeedMin())
                {
                    GetInteroperability().CloseAll();
                    (*((*m_Params).ReturnPastBalances())).EmptyProfits();
                    GetLogger().LogSpecial(IString("Closing All Orders and Reversing, Balance : ") + GetInteroperability().DoubleToString(balance) + ", Speed : " + GetInteroperability().DoubleToString(speed));
                    m_ReverseTradesAndLotSize.m_Bool = !m_ReverseTradesAndLotSize.m_Bool;
                }
            }
        }

		void CheckForCloseAll(double prof)
		{
			if (prof < m_MaxProfitValue - (*CTrading::GetVaribles()).MultivariateTrailStop())
			{
				GetInteroperability().CloseAll();
				m_MaxProfitValue = 0;
				GetLogger().LogSpecial(IString("Closing All Orders, Profit : ") + GetInteroperability().DoubleToString(prof));
			}
			else
			{
				if (prof > m_MaxProfitValue)
				{
					m_MaxProfitValue = prof;
				}
			}
			if (CConfig::MultivariateTrailStopMinimum)
			{
				if (prof > m_MinProfitValue + (*CTrading::GetVaribles()).MultivariateTrailStop())
				{
					GetInteroperability().CloseAll();
					m_MinProfitValue = 0;
					GetLogger().LogSpecial(IString("Closing All Orders, Profit : ") + GetInteroperability().DoubleToString(prof));
				}
				else
				{
					if (prof < m_MinProfitValue)
					{
						m_MinProfitValue = prof;
					}
				}
			}
		}

		void CheckForCloseAllBySpeed(double prof, IDateTime &date)
		{
            (*((*m_Params).ReturnPastProfits())).AddProfit(prof, date, (*CTrading::GetVaribles()).MultivariateTrailSpeedHours());
			if ((*((*m_Params).ReturnPastProfits())).IsValid((*CTrading::GetVaribles()).MultivariateTrailSpeedHours()))
			{
				if ((*CTrading::GetVaribles()).MultivariateTrailSpeed() > (prof - (*((*m_Params).ReturnPastProfits())).PastProfit()) / (*CTrading::GetVaribles()).MultivariateTrailSpeedHours())
				{
					if (prof != 0.0)
					{
						GetLogger().LogSpecial(IString("Closing All."));
						GetInteroperability().CloseAll();
                        (*((*m_Params).ReturnPastProfits())).EmptyProfits();
					}
				}
			}
		}

		void CheckForCloseProfits(IGenericArray<double> &profits, IGenericArray<int> &numbers, IDateTime &date)
		{
			double profit = 0.0;
			double reverseProfit = 0.0;

            (*((*m_Params).ReturnOpenTrades())).RemoveNotPresent(numbers);

			for (int i = 0; i < numbers.Size(); i++)
			{
				if ((*((*m_Params).ReturnOpenTrades())).IsReverseTrade(numbers.GetValue(i)))
				{
					reverseProfit += profits.GetValue(i);
				}
				else
				{
					profit += profits.GetValue(i);
				}
			}

			if (profit < (*CTrading::GetVaribles()).MultivariateProfitStopMin())
			{
				GetLogger().LogSpecial(IString("Closing Opened Profit Min."));
				GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnOpenTrades());
			}
			if (profit > (*CTrading::GetVaribles()).MultivariateProfitStopMax())
			{
				GetLogger().LogSpecial(IString("Closing Opened Profit Max."));
				GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnOpenTrades());
			}
			if (reverseProfit < (*CTrading::GetVaribles()).MultivariateReverseProfitStopMin())
			{
				GetLogger().LogSpecial(IString("Closing Opened Reverse Profit Min."));
				GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnReverseOpenTrades());
			}
			if (reverseProfit > (*CTrading::GetVaribles()).MultivariateReverseProfitStopMax())
			{
				GetLogger().LogSpecial(IString("Closing Opened Reverse Profit Max."));
				GetInteroperability().CloseAll((*((*m_Params).ReturnOpenTrades())).ReturnReverseOpenTrades());
			}
		}

    public:
		void InitFarPastData()
		{
			if (CConfig::WriteFarPastData)
			{
				CFarPastDataFile::WriteDataColumnsNames(*m_Pairs);
				m_LastFarPastDate = IDateTime::TimeCurrent();
			}
			if (CConfig::WritePastProfit)
			{
				m_LastPastProfit = IDateTime::TimeCurrent();
			}
		}

        void IntervalElapsedFarPastData()
		{
			if (CConfig::WriteFarPastData)
			{
				IDateTime date = IDateTime::TimeCurrent();

				if (IDateTime::SecondsDiff(m_LastFarPastDate, date) > CConfig::FarPastDataIntervalSeconds)
				{
					CFarPastDataFile::WriteDataColumnsValues(*m_Pairs);
					m_LastFarPastDate = date;
				}
			}
		}

		void IntervalElapsedPastProfit()
		{
			if (CConfig::WritePastProfit || CConfig::ConditionMultivariateTrailStop != MultivariateTrailStop::NONE)
			{
				IDateTime date = IDateTime::TimeCurrent();

				if (IDateTime::SecondsDiff(m_LastPastProfit, date) >= CConfig::PastProfitIntervalSeconds)
				{
					double prof;
					IGenericArray<double> profits;
					IGenericArray<int> numbers;

					prof = GetInteroperability().ProfitAll(numbers, profits);
					
					if (CConfig::WritePastProfit)
					{
						double balance;
						double lot;
						IString res;

						balance = GetInteroperability().AccountBalance();
						lot = (*CTrading::GetVaribles()).LotSize();

						for (int i = 0; i < numbers.Size(); i++)
						{
							if (!(*((*m_Params).ReturnOpenTrades())).IsReverseTrade(numbers.GetValue(i)))
							{
								res += (IString(";") + GetInteroperability().IntegerToString(numbers.GetValue(i)) + IString(":") + GetInteroperability().DoubleToString(profits.GetValue(i)) + IString(",false"));
							}
							else
							{
								res += (IString(";") + GetInteroperability().IntegerToString(numbers.GetValue(i)) + IString(":") + GetInteroperability().DoubleToString(profits.GetValue(i)) + IString(",true"));
							}
						}

						GetLogger().LogFromProfitPast(IString("Profit : ") + GetInteroperability().DoubleToString(prof) + IString(", Account Balance : ") + GetInteroperability().DoubleToString(balance) + IString(", Lot Size : ") + GetInteroperability().DoubleToString(lot) + IString(", Profits : ") + res);
					}

					m_LastPastProfit = date;
                    m_ReverseTradesAndLotSize.m_Double = (*CTrading::GetVaribles()).LotSize();

                    if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::TRAIL_STOP)
					{
						CheckForCloseAll(prof);
					}
					else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SPEED_STOP)
					{
						CheckForCloseAllBySpeed(prof, date);
					}
					else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_STOP)
					{
						CheckForClosePart(profits, numbers, date);
					}
					else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_PROFIT)
					{
						CheckForCloseProfits(profits, numbers, date);
					}
					else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_REVERSE_PROFITS)
					{
						CheckForReverse(prof);
					}
					else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_REVERSE_BALANCE)
					{
						CheckForReverse();
					}
                    else if (CConfig::ConditionMultivariateTrailStop == MultivariateTrailStop::SURFING_REVERSE_BALANCE_SPEED)
                    {
                        CheckForReverseBalanceSpeed(date);
                    }
				}
			}
		}

		void SetParams(CParams &params, CCurrencyPairs &pairs)
        {
            m_Params = &params;
            m_Pairs = &pairs;
			CValueAnalyzer::SetParams(params);
        }
    
		bool LogSpeeds()
		{
			m_TrailSpeedLogCountDown--;
			if (m_TrailSpeedLogCountDown == 0)
			{
				m_TrailSpeedLogCountDown = CConfig::TrailSpeedLogCountDown;
				return true;
			}
			return false;
		}
		
		void SetInitialization()
        {
            m_LastInitialization = IDateTime::TimeCurrent();
        }

        void AssignAnalyzerProcess(CMultiAnalyzerProcess &process)
        {
			m_BalanceLastMax = GetInteroperability().AccountBalance();
            m_AnalyzerProcess = &process;
			(*m_AnalyzerProcess).SetPastTradesAnalyzer(m_PastTradesAnalyzer, m_PastTradesLearn);
			(*m_AnalyzerProcess).AssignDistancesProvider(m_DistancesProvider);
			(*m_AnalyzerProcess).AssignOpenTradesObject(*((*m_Params).ReturnOpenTrades()));
			(*m_AnalyzerProcess).AssignReverseTradesMethodG(m_ReverseTradesAndLotSize);
        }
        
        void KillTimer()
        {
            m_Timer.KillTimer();
        }
        
        void SetTimer()
        {
            m_Timer.SetTimer();
        }
    
        static bool CheckLeverage()
        {
            GetLogger().LogSpecial(IString("Account Leverage equal to : ") + GetInteroperability().IntegerToString(GetInteroperability().AccountLeverage()));
            if(CConfig::CheckLeverage)
            {
                if(GetInteroperability().AccountLeverage() > MaxLeverage)
                {
                    GetLogger().LogSpecial(IString("Account Leverage can be at max : ") + GetInteroperability().IntegerToString(MaxLeverage));
                    return false;
                }
            }
            return true;
        }
        
        bool CheckForInitialization()
        {
            IDateTime date = IDateTime::TimeCurrent();
            
            if(IDateTime::MinutesDiff(m_LastInitialization, date) > CConfig::FutureHoursWindow * 60)
            {
                return true;
            }
            else if(!IsValid())
            {
                if(IDateTime::MinutesDiff(m_LastInitialization, date) > CConfig::FutureHoursInvalidRestart * 60)
                {
                    return true;
                }
            }
            return false;
        }
        
        CCalculusRobotExpert()
        {
     	    IString logs("Logs");

            GetLogger().Init(logs);
            CLearning::InitStatic();
			m_TrailSpeedLogCountDown = CConfig::TrailSpeedLogCountDown;
			m_ReverseTradesAndLotSize.m_Bool = false;
            m_ReverseTradesAndLotSize.m_Double = CConfig::LotSize;
	    }
      
        bool GetValues()
        {
            IDateTime now = IDateTime::TimeCurrent();
            
            for(int i=0;i<(*m_Pairs).AmountCurrencyPairsToBuy();i++)
            {
				IString pair((*m_Pairs).ReturnCurrencyPairToBuyNr(i));

                GetLogger().Log(IString((*m_Pairs).ReturnCurrencyPairToBuyNr(i)) + " : " + GetInteroperability().DoubleToString(CValues::GetPriceAtTime(pair, now)));
            }

            for(int i=0;i<(*m_Pairs).AmountCurrencyPairsToSell();i++)
            {
				IString pair((*m_Pairs).ReturnCurrencyPairToSellNr(i));

                GetLogger().Log(IString((*m_Pairs).ReturnCurrencyPairToSellNr(i)) + " : " + GetInteroperability().DoubleToString(CValues::GetPriceAtTime(pair, now)));
            }
            
            for(int i=0;i<(*m_Pairs).AmountOtherSymbolsToBuy();i++)
            {
				IString pair((*m_Pairs).ReturnOtherSymbolToBuyNr(i));

                GetLogger().Log(IString((*m_Pairs).ReturnOtherSymbolToBuyNr(i)) + " : " + GetInteroperability().DoubleToString(CValues::GetPriceAtTime(pair, now)));
            }

            for(int i=0;i<(*m_Pairs).AmountOtherSymbolsToSell();i++)
            {
				IString pair((*m_Pairs).ReturnOtherSymbolToSellNr(i));

                GetLogger().Log(IString((*m_Pairs).ReturnOtherSymbolToSellNr(i)) + " : " + GetInteroperability().DoubleToString(CValues::GetPriceAtTime(pair, now)));
            }
            
            if(!m_Values.InitWithNowTime(*m_Pairs))
            {
                return false;
            }
            else
            {
                m_Values.Log();
                return true;
            }
        } 
         
        
        CValues* ReturnValues()
        {
            return &m_Values;
        }
         
		void CheckHistory()
		{
			m_AnlyzerInit.PreInit(*m_Pairs);
			m_AnlyzerInit.SetValues(m_Values);

			m_AnlyzerInit.CheckHistory();

			if (CConfig::UseTrendFromLogs && (*m_Pairs).TradablePeriod())
			{
				TrainTrendNetwork();
			}
			if (CConfig::UseAfterTrend && CConfig::UseAfterTrendAutomaticParams && (*m_Pairs).TradablePeriod())
			{
				DetermineMethodDPArams();
			}
			if (CConfig::UseSellBuyNeuralNetworks && (*m_Pairs).TradablePeriod())
			{
				TrainNeuralNetworks();
	        }
			if (CConfig::UseIndividualNeuralNetworks && (*m_Pairs).TradablePeriod())
			{
     			LearnSymbolNets();
			}
			if (CConfig::UseCalculateDistance && (*m_Pairs).TradablePeriod())
			{
				AnalyzePastData();
			}
		}

		void TrainTrendNetwork()
		{
			if (CConfig::UseTrendFromLogs)
			{
			    CLogs logs;

			    logs.ReadLogs();
			    logs.ExtractTrades((*m_Pairs).TradableAmount());
				logs.AssignOrder();

			    GetLogger().LogSpecial(IString("Amount of Trades Extracted from logs : ") + GetInteroperability().IntegerToString(logs.TradesAmount()));
			    if (logs.TradesAmount() > CConfig::TradesAmountMin)
			    {
				    GetLogger().LogSpecial(IString("Detecting Trend."));
			    }
			    else
			    {
				    GetLogger().LogSpecial(IString("Detecting Trend Skipped. Extracted Trades Amount is less then : ") + GetInteroperability().IntegerToString(CConfig::TradesAmountMin));
					return;
			    }

				m_PastTradesAnalyzer.Analyze(logs);
				m_PastTradesAnalyzer.CleanUpAndSetMultiplicationValues();

				if (CConfig::PastTrendsUseNeuralNetwork)
				{
					m_PastTradesLearn.Init(m_PastTradesAnalyzer.BestInputsAmount(), 1, CConfig::HiddenLayersAmountForTrendNetwork, CConfig::ActivationFunctionDimensionForTrendNetwork);
					logs.AssignVectors(m_PastTradesAnalyzer.BestInputsAmount());
					m_PastTradesLearn.Learn(*(logs.ReturnVectors()), 0.0);
				}
			}
		}
        
        void TrainNeuralNetworks()
        {
            IGenericObjectArray<CLineProcessed> differentialLinesBuys;
            IGenericObjectArray<CLineProcessed> differentialLinesSells;
        
            CLinesProcessor::GenerateDifferentialLines(m_AnlyzerInit.PastPointsBuys, differentialLinesBuys, m_Values);
            CLinesProcessor::GenerateDifferentialLines(m_AnlyzerInit.PastPointsSells, differentialLinesSells, m_Values);
        
            CVectors vectorsBuy;
            CVectors vectorsSell;
            
            int pos = 0;
            double val, diff;
            int SymbolNr;
                    
            if(differentialLinesBuys.Size() > 0)
            {
                vectorsBuy.Init(differentialLinesBuys.Size(), (*(differentialLinesBuys.GetPointerToValue(0))).Points.Size() + 1 + 1, 1);
                           
                for(int i=0;i<m_AnlyzerInit.PastPointsBuys.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).IsValid())
                    {     
                        SymbolNr = (*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).SymbolNr;
                        for(int j=0;j<(*(differentialLinesBuys.GetPointerToValue(0))).Points.Size();j++)
                        {
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Input.SetValue(j, (*(differentialLinesBuys.GetPointerToValue(pos))).Points.GetValue(j));
                        }
                        if((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).Successful)
                        {
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Output.SetValue(0, 1.0);
                        }
                        else
                        {                    
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Output.SetValue(0, -1.0);
                        }
                        if((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).Symbol)
                        {
                            val = m_Values.ReturnPastValueNrToBuy((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                            diff = GetInteroperability().MathAbs(val - m_Values.ReturnPastValueNrToBuy((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesBuys.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, true, true));
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesBuys.GetPointerToValue(0))).Points.Size() + 1,  (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, true, true));
                        }
                        else
                        {
                            val = m_Values.ReturnOtherPastValueNrToBuy((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                            diff = GetInteroperability().MathAbs(val - m_Values.ReturnOtherPastValueNrToBuy((*(m_AnlyzerInit.PastPointsBuys.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesBuys.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, false, true));
                            (*(vectorsBuy.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesBuys.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, false, true));
                        }
                        pos++;
                    }
                }
            }
            
            if(differentialLinesSells.Size() > 0)
            {
                vectorsSell.Init(differentialLinesSells.Size(), (*(differentialLinesSells.GetPointerToValue(0))).Points.Size() + 1 + 1, 1);
        
                pos = 0;
                for(int i=0;i<m_AnlyzerInit.PastPointsSells.Size();i++)
                {
                    if((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).IsValid())
                    {     
                        SymbolNr = (*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).SymbolNr;
                        for(int j=0;j<(*(differentialLinesSells.GetPointerToValue(0))).Points.Size();j++)
                        {
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Input.SetValue(j, (*(differentialLinesSells.GetPointerToValue(pos))).Points.GetValue(j));
                        }
                        if((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).Successful)
                        {
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Output.SetValue(0, 1.0);
                        }
                        else
                        {                    
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Output.SetValue(0, -1.0);
                        }
                        if((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).Symbol)
                        {
                            val = m_Values.ReturnPastValueNrToSell((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                            diff = GetInteroperability().MathAbs(val - m_Values.ReturnPastValueNrToSell((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesSells.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, true, false));
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesSells.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, true, false));
                        }
                        else
                        {
                            val = m_Values.ReturnOtherPastValueNrToSell((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).EstablishingMinimumMaximumPointNr, SymbolNr);
                            diff = GetInteroperability().MathAbs(val - m_Values.ReturnOtherPastValueNrToSell((*(m_AnlyzerInit.PastPointsSells.GetPointerToValue(i))).MinimumMaximumPointNr, SymbolNr));
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesSells.GetPointerToValue(0))).Points.Size(), (*m_Pairs).ReturnUpperLimit(val, diff, SymbolNr, false, false));
                            (*(vectorsSell.Vectors.GetPointerToValue(pos))).Input.SetValue((*(differentialLinesSells.GetPointerToValue(0))).Points.Size() + 1, (*m_Pairs).ReturnLowerLimit(val, diff, SymbolNr, false, false));
                        }
                        pos++;
                    }
                }
            }
            
            if(differentialLinesBuys.Size() > 0)
            {
                if(LearnTestNetForBuy(differentialLinesBuys, vectorsBuy))
                {
                    LearnNetForBuy(differentialLinesBuys, vectorsBuy);
                }
            }
            
            if(differentialLinesSells.Size() > 0)
            {   
                if(LearnTestNetForSell(differentialLinesSells, vectorsSell))
                {
                    LearnNetForSell(differentialLinesSells, vectorsSell);
                }
            }
            (*m_AnalyzerProcess).AssignNetworks(m_NetBuys, m_NetSells);
        }
        
        void DoTrading()
        {
            if(CConfig::CheckBalance && GetInteroperability().AccountBalance() >= 200000.0)
            {
				GetLogger().Log(IString("Account Balance reached maximum."));
                return;
            }
            else
            {
          	   IDateTime time(IDateTime::TimeCurrent());

               (*m_AnalyzerProcess).AddToAnalysis(time);
            }
        }
};
const int CCalculusRobotExpert::MaxLeverage = 30;
//+------------------------------------------------------------------+
#endif
