/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

#ifndef FILE_CALCULUS_ROBOT_RUNNER
#define FILE_CALCULUS_ROBOT_RUNNER

#include "CalculusRobotExpert.h"
#include "LibrariesInteroperability.h"
#include "LibrariesLogger.h"
#include "CR Includes/Logs.h"
#include "CR Includes/Varibles.h"
#include "CR Includes/Rules.h"

//+------------------------------------------------------------------+
class CCalculusRobotRunner {
private:
	CCalculusRobotExpert *Expert;
	CParams Params;
	CCurrencyPairs Pairs;
	CMultiAnalyzerProcess AnalyzerProcess;

public:
	bool Init()
	{
		GetInteroperability().Activate();
		CFarPastDataFile::Init();
        CVaribles::Init();
		CConfig::Init();
		CRules::Init();

		if (!CCalculusRobotExpert::CheckLeverage())
		{
			return false;
		}

		Expert = new CCalculusRobotExpert();
		(*Expert).SetTimer();

		if (CConfig::ConsiderAccountCurrency)
		{
			IString curr(GetInteroperability().AccountCurrency());

			Pairs.Init(curr);
		}
		else
		{
			IString curr("000");

			Pairs.Init(curr);
		}

		(*Expert).SetParams(Params, Pairs);

		if (!(*Expert).GetValues())
		{
			GetLogger().Log(IString("Can't get history. Re-run Robot."));
			return false;
		}
		else
		{
			IDateTime time(IDateTime::TimeCurrent());
			CValues *values = (*Expert).ReturnValues();

			AnalyzerProcess.PreInit(Pairs);
			AnalyzerProcess.SetValues(*values);
			AnalyzerProcess.Init(Pairs, time, Params);
			(*Expert).AssignAnalyzerProcess(AnalyzerProcess);
			(*Expert).CheckHistory();
			(*Expert).SetInitialization();
			(*Expert).InitFarPastData();
		}
		return true;
	}

	void OnInterval()
	{
		if ((*Expert).CheckForInitialization())
		{
			delete Expert;
			Expert = new CCalculusRobotExpert();

			(*Expert).SetParams(Params, Pairs);
			AnalyzerProcess.ZeroSearchingProfit();

			if (!(*Expert).GetValues())
			{
				GetLogger().Log(IString("Can't get history. Re-run Robot."));
			}
			(*Expert).AssignAnalyzerProcess(AnalyzerProcess);

			(*Expert).CheckHistory();
			(*Expert).SetInitialization();
			(*Expert).InitFarPastData();
		}
		else
		{
			(*Expert).DoTrading();
			(*Expert).IntervalElapsedFarPastData();
			(*Expert).IntervalElapsedPastProfit();
		}
	}

	void DeInit()
	{
		if (Expert != NULL)
		{
			(*Expert).KillTimer();
			delete Expert;
		}
	}
};
//+------------------------------------------------------------------+
#endif