/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

#ifndef FILE_INTEROPERABILIT
#define FILE_INTEROPERABILIT

#ifndef CPP

#include "String.h"
#include "DateTime.h"
#include "GenericArray.h"
#include "SortableArray.h"

//+------------------------------------------------------------------+
class IInteroperability
{
    public:
		void Activate()
		{
#ifdef MQL4  
			::RefreshRates();
#endif             
#ifdef MQL5      
            
#endif
		}

        void Print(IString &str)
        {
            ::Print(str.GetPlatformString());
        }        
    
        void RefreshRates()
        {
#ifdef MQL4           
            ::RefreshRates();
#endif             
#ifdef MQL5      
            
#endif
        }  
        
        int AccountLeverage()
        {
#ifdef MQL4           
           return ::AccountLeverage();
#endif             
#ifdef MQL5      
           return (int)::AccountInfoInteger(ACCOUNT_LEVERAGE);
#endif
        }  
        
		unsigned long GetLastError()
		{
			return ::GetLastError();
		}

        IString AccountCurrency()
        {
#ifdef MQL4           
           return IString(::AccountCurrency());
#endif             
#ifdef MQL5      
           return IString(::AccountInfoString(ACCOUNT_CURRENCY));
#endif
        }
        
        bool OrderSelect(int nr, ulong ticket)
        {
#ifdef MQL4           
           return ::OrderSelect(nr,SELECT_BY_POS,MODE_HISTORY);
#endif             
#ifdef MQL5      
           return ::HistoryOrderSelect(ticket);
#endif
        }

		double OrderTakeProfit(ulong ticket)
		{
#ifdef MQL4           
			return ::OrderTakeProfit();
#endif             
#ifdef MQL5      
			return ::HistoryOrderGetDouble(ticket, ORDER_TP);
#endif
		}

		double OrderStopLoss(ulong ticket)
		{
#ifdef MQL4           
			return ::OrderStopLoss();
#endif             
#ifdef MQL5      
			return ::HistoryOrderGetDouble(ticket, ORDER_SL);
#endif
		}

		bool OrderClose(ulong ticket)
		{
#ifdef MQL4           
			if (::OrderType() == OP_BUY)
			{
				return ::OrderClose(::OrderTicket(), ::OrderLots(), ::MarketInfo(::OrderSymbol(), MODE_BID), CConfig::SlipPage, Red);
			}
			else
			{
				return ::OrderClose(::OrderTicket(), ::OrderLots(), ::MarketInfo(::OrderSymbol(), MODE_ASK), CConfig::SlipPage, Red);
			}
#endif             
#ifdef MQL5
			IString msg("Order Close not implemented.");

			Print(msg);

			return true;
#endif
		}

        unsigned long HistoryOrderTicket(int nr)
        {
#ifdef MQL4           
           return 0;
#endif             
#ifdef MQL5      
           return ::HistoryOrderGetTicket(nr);
#endif
        }
        
        void OrderOpenTime(ulong ticket, IDateTime &res)
        {
#ifdef MQL4           
           res = IDateTime(::OrderOpenTime());
#endif             
#ifdef MQL5      
           res = IDateTime(::HistoryOrderGetInteger(ticket, ORDER_TIME_DONE));
#endif
        }
        
        double OrderOpenPrice(ulong ticket)
        {
#ifdef MQL4           
           return ::OrderOpenPrice();
#endif             
#ifdef MQL5      
           return ::HistoryOrderGetDouble(ticket, ORDER_PRICE_OPEN);
#endif
        }

        double OrderProfit(ulong ticket)
        {
#ifdef MQL4           
           return ::OrderProfit();
#endif             
#ifdef MQL5      
           return ::HistoryDealGetDouble(ticket, DEAL_PROFIT);
#endif
        }

        double OrderCurrentPrice(ulong ticket)
        {
            IString symbol;
            
            OrderSymbol(ticket, symbol);
#ifdef MQL4           
            return BuyPrice(symbol);           
#endif             
#ifdef MQL5      
            return BuyPrice(symbol);
#endif
        }

        void OrderSymbol(ulong ticket, IString &res)
        {
#ifdef MQL4           
           res = IString(::OrderSymbol());
#endif             
#ifdef MQL5      
           res = IString(::HistoryOrderGetString(ticket, ORDER_SYMBOL));
#endif
        }

        double AccountBalance()
        {
#ifdef MQL4           
           return ::AccountBalance();
#endif             
#ifdef MQL5      
           return ::AccountInfoDouble(ACCOUNT_BALANCE);
#endif
        }  
        
        double GetPriceAtTime(IString &symbol, IDateTime &date)
        { 
#ifdef MQL4 
            int iWhenM1 = iBarShift(symbol.GetPlatformString(), PERIOD_M1, date.GetPlatformTime());
            return iLow(symbol.GetPlatformString(), PERIOD_M1, iWhenM1);
#endif             
#ifdef MQL5      
            int iWhenM1 = iBarShift(symbol.GetPlatformString(), PERIOD_M1, date.GetPlatformTime());
            return iLow(symbol.GetPlatformString(), PERIOD_M1, iWhenM1);
#endif
        }
        
        double GetPriceAtTimeHigh(IString &symbol, IDateTime &date)
        { 
#ifdef MQL4 
            int iWhenM1 = iBarShift(symbol.GetPlatformString(), PERIOD_M1, date.GetPlatformTime());
            return iHigh(symbol.GetPlatformString(), PERIOD_M1, iWhenM1);
#endif             
#ifdef MQL5      
            int iWhenM1 = iBarShift(symbol.GetPlatformString(), PERIOD_M1, date.GetPlatformTime());
            return iHigh(symbol.GetPlatformString(), PERIOD_M1, iWhenM1);
#endif
        }
        
        double Spread(IString &symbol)
        {
#ifdef MQL4 
            return MarketInfo(symbol.GetPlatformString(), MODE_ASK) - MarketInfo(symbol.GetPlatformString(), MODE_BID);
#endif             
#ifdef MQL5      
            return SymbolInfoDouble(symbol.GetPlatformString(), SYMBOL_ASK) - SymbolInfoDouble(symbol.GetPlatformString(), SYMBOL_BID);
#endif
        }
        
        double BuyPrice(IString &symbol)
        {
#ifdef MQL4 
            return MarketInfo(symbol.GetPlatformString(), MODE_ASK);
#endif             
#ifdef MQL5      
            return ::SymbolInfoDouble(symbol.GetPlatformString(),SYMBOL_ASK);
#endif
        }
        
        double SellPrice(IString &symbol)
        {
#ifdef MQL4 
            return MarketInfo(symbol.GetPlatformString(), MODE_BID);
#endif             
#ifdef MQL5      
            return ::SymbolInfoDouble(symbol.GetPlatformString(), SYMBOL_BID);
#endif
        }
        
        int OrderSend(IString &symbol, bool buy, double lotSize, double price, int slippage, double stopLoss, double takeProfit)
        {
#ifdef MQL4 
            return ::OrderSend(symbol.GetPlatformString(), buy ? OP_BUY : OP_SELL, lotSize, price, slippage, stopLoss, takeProfit);
#endif             
#ifdef MQL5      
            MqlTradeResult result;
            MqlTradeRequest request;
            
            ZeroMemory(request);
            ZeroMemory(result);
            request.symbol = symbol.GetPlatformString();
            request.action = TRADE_ACTION_DEAL;
            request.type_filling = ORDER_FILLING_FOK;
            request.type = buy ? ORDER_TYPE_BUY : ORDER_TYPE_SELL;
            request.volume = lotSize;
            request.price = price;
            request.sl = stopLoss;
            request.tp = takeProfit;
            request.deviation = slippage;
            
            if(::OrderSend(request, result))
            {
                return result.request_id;
            }
            else
            {
                return -1;
            }        
#endif
        }  

        int OrdersHistoryTotal()
        {
#ifdef MQL4 
            return ::OrdersHistoryTotal();
#endif             
#ifdef MQL5      
            return ::HistoryOrdersTotal();
#endif
        }

		static IDateTime StringToTime(IString &str)
		{
#ifdef MQL4 
			return IDateTime(::StrToTime(str.GetPlatformString()));
#endif             
#ifdef MQL5      
			return IDateTime(::StringToTime(str.GetPlatformString()));
#endif
		}

		bool CloseAll()
		{
#ifdef MQL4
			bool success = true;

			for (int x = ::OrdersTotal(); x >= 0; x--)
			{
				if (::OrderSelect(x, SELECT_BY_POS) == true)
				{
					if (::OrderClose(::OrderTicket(), ::OrderLots(), ::OrderClosePrice(), CConfig::SlipPage, CLR_NONE))
					{
						
					}
					else
					{
						success = false;
					}
				}
			}
			return success;
#endif
#ifdef MQL5
			IString msg("Close All not implemented.");

			Print(msg);

            return false; 
#endif
		}

		bool CloseAll(ISortableArray<int> *numbers)
		{
#ifdef MQL4
			bool success = true;

			for (int x = ::OrdersTotal(); x >= 0; x--)
			{
				if (::OrderSelect(x, SELECT_BY_POS) == true)
				{
					int tradeNr = (*numbers).BinarySearch(::OrderTicket());

					if (tradeNr < 0 || tradeNr >= (*numbers).Size())
					{
						return false;
					}

					if (::OrderTicket() == (*numbers).GetPos(tradeNr))
					{
						if (::OrderClose(::OrderTicket(), ::OrderLots(), ::OrderClosePrice(), CConfig::SlipPage, CLR_NONE))
						{

						}
						else
						{
							success = false;
						}
					}
				}
			}
			return success;
#endif
#ifdef MQL5
			IString msg("Close All not implemented.");

			Print(msg);

			return false;
#endif
		}

		double ProfitAll(IGenericArray<int> &numbers, IGenericArray<double> &profits)
		{
#ifdef MQL4
			double profit = 0.0;
			int total = ::OrdersTotal();

			numbers.Resize(total);
			profits.Resize(total);

			for (int x = total - 1 ; x >= 0; x--)
			{
				if (::OrderSelect(x, SELECT_BY_POS) == true)
				{
					profit += ::OrderProfit();
					numbers.SetValue(x, ::OrderTicket());
					profits.SetValue(x, ::OrderProfit());
				}
			}
			return profit;
#endif
#ifdef MQL5
			int total = OrdersHistoryTotal();
			IDateTime dateOrder;
			unsigned long ticket;
			double totalProf = 0.0;

			numbers.Resize(total);
			profits.Resize(total);

			for (int i = 0; i < total; i++)
			{
				ticket = HistoryOrderTicket(i);
				if (OrderSelect(i, ticket) == true)
				{
					totalProf += OrderProfit(ticket);
					numbers.SetValue(i, ticket);
					profits.SetValue(i, OrderProfit(ticket));
				}
			}
			return totalProf;
#endif
		}

		IString TimeToString(IDateTime &time)
		{
			IString str;
			string tm = ::TimeToString(time.GetPlatformTime());

			str.AssignString(tm);
			return str;
		}

      int SymbolsTotal(bool shown)
      {
          return ::SymbolsTotal(shown);
      }

      IString SymbolName(int nr, bool shown)
      {
          IString symbol(::SymbolName(nr, shown));
          
          return symbol;
      }

		double MathAbs(double val)
		{
			return ::MathAbs(val);
		}

		double MathExp(double val)
		{
			return ::MathExp(val);
		}

		void EventSetTimer(int secs)
		{
			::EventSetTimer(secs);
		}

		void EventKillTimer()
		{
			::EventKillTimer();
		}

		bool MathIsValidNumber(double val)
		{
			return ::MathIsValidNumber(val);
		}

		int MathCeil(double val)
		{
			return (int)::MathCeil(val);
		}

		void MathSrand()
		{
			::MathSrand(::GetTickCount());
		}

		int MathRand()
		{
			return ::MathRand();
		}
		
		IString DoubleToString(double val)
		{
		    return IString(::DoubleToString(val));
		}

		IString BoolToString(bool val)
		{
			if (val)
			{
				return IString("true");
			}
			else
			{
				return IString("false");
			}
		}
		
		static IString IntegerToString(long val)
		{
		    return IString(::IntegerToString(val));
		}

		double NormalizeDouble(double price, int digits)
		{
			return ::NormalizeDouble(price, digits);
		}

		int Digits()
		{
			return ::Digits();
		}

		double StringToDouble(IString &str)
		{
			return ::StringToDouble(str.GetPlatformString());
		}
};
//+------------------------------------------------------------------+
#else

#include "../CPP/CalculusRobot/Internal/InternalCppInteroperability.h"

#endif
#endif
//+------------------------------------------------------------------+
