﻿/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using System;
using System.Collections.Generic;

namespace CommonLibrary
{
    public class LookSimulator
    {
        private List<Look> Looks { get; set; }
        private SortedList<int, int> RemovedProfits { get; set; }
        private SortedList<int, int> ReversedProfits { get; set; }

        public void Clear()
        {
            Looks.Clear();
            RemovedProfits.Clear();
            ReversedProfits.Clear();
        }

        public void ClearRemovedProfits()
        {
            RemovedProfits.Clear();
            ReversedProfits.Clear();
        }

        public LookSimulator()
        {
            RemovedProfits = new SortedList<int, int>();
            ReversedProfits = new SortedList<int, int>();
        }

        public void SetLooks(FileLooks fl)
        {
            Looks = fl.Looks;
        }

        public void SetLooks(List<Look> looks)
        {
            Looks = looks;
        }

        private bool GetPastLook(int lookNr, double hours, out int pastLookNr)
        {
            pastLookNr = 0;
            for (int i = lookNr; i >= 0; i--)
            {
                if ((Looks[lookNr].Date - Looks[i].Date).TotalMinutes >= hours * 60.0)
                {
                    pastLookNr = i;
                    return true;
                }
            }
            return false;
        }

        private bool GetSpeedAtLook(int lookNr, double hours, out double speed)
        {
            int pastLookNr;

            speed = 0.0;
            if (GetPastLook(lookNr, hours, out pastLookNr))
            {
                speed = (ProfitAtLook(lookNr) - ProfitAtLook(pastLookNr)) / (Looks[lookNr].Date - Looks[pastLookNr].Date).TotalHours;
                return true;
            }
            else
            {
                return false;
            }
        }

        public void MinMaxSpeed(double hours, out double minSpeed, out double maxSpeed)
        {
            double speed = 0.0;

            minSpeed = 0.0;
            maxSpeed = 0.0;

            for (int i = 0; i < Looks.Count; i++)
            {
                if (GetSpeedAtLook(i, hours, out speed))
                {
                    if (speed > maxSpeed)
                    {
                        maxSpeed = speed;
                    }
                    if (speed < minSpeed)
                    {
                        minSpeed = speed;
                    }
                }
            }
        }

        private int LooksAmount()
        {
            return Looks.Count;
        }

        private double ProfitAtLook(int nr)
        {
            double profits = 0.0;

            foreach (LookProfit lk in Looks[nr].Profits)
            {
                if (!(RemovedProfits.ContainsValue(lk.ProfitId)))
                {
                    if (ReversedProfits.ContainsValue(lk.ProfitId))
                    {
                        profits -= lk.Profit;
                    }
                    else
                    {
                        profits += lk.Profit;
                    }
                }
            }
            return profits;
        }

        private bool RemoveProfit(int nr)
        {
            if (!RemovedProfits.ContainsKey(nr))
            {
                RemovedProfits.Add(nr, nr);
                return true;
            }
            return false;
        }

        private bool ClosedProfit(int lookCurrent, int profitId)
        {
            if (RemovedProfits.ContainsKey(profitId))
            {
                return false;
            }
            foreach (LookProfit l in Looks[lookCurrent].Profits)
            {
                if (l.ProfitId == profitId)
                {
                    return false;
                }
            }
            return true;
        }

        private int RemoveProfits(int lookNr)
        {
            int removedAm = 0;

            foreach (LookProfit lp in Looks[lookNr].Profits)
            {
                if (RemoveProfit(lp.ProfitId))
                {
                    removedAm++;
                }
            }
            return removedAm;
        }

        public double CalculateProfit(double diff, out int posAmount, out int negAmount, bool nmts = false)
        {
            double max = 0.0;
            double profit = 0.0;
            double tmp;
            double min = 0.0;

            RemovedProfits.Clear();
            posAmount = 0;
            negAmount = 0;

            for (int i = 0; i < LooksAmount(); i++)
            {
                tmp = ProfitAtLook(i);
                if (tmp < max - diff)
                {
                    if (tmp >= 0.0)
                    {
                        posAmount++;
                    }
                    else
                    {
                        negAmount++;
                    }
                    profit += tmp;
                    RemoveProfits(i);
                    max = 0.0;
                }
                else
                {
                    if (tmp > max)
                    {
                        max = tmp;
                    }
                }
                if (nmts)
                {
                    if (tmp > min + diff)
                    {
                        if (tmp >= 0.0)
                        {
                            posAmount++;
                        }
                        else
                        {
                            negAmount++;
                        }
                        profit += tmp;
                        RemoveProfits(i);
                        min = 0.0;
                    }
                    else
                    {
                        if (tmp < min)
                        {
                            min = tmp;
                        }
                    }
                }
                if (i > 0)
                {
                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                            profit += (lp.Profit);
                        }
                    }
                }
            }
            return profit;
        }

        public double CalculateProfitWithSpeed(double speed, double hours, out int ClosedAmount)
        {
            double profit = 0.0;
            double tmp;
            PastProfits pp = new PastProfits(hours);
            ClosedAmount = 0;

            RemovedProfits.Clear();

            for (int i = 0; i < LooksAmount(); i++)
            {
                tmp = ProfitAtLook(i);
                pp.AddProfit(tmp, Looks[i].Date);
                if (pp.IsValid)
                {
                    if (speed > (tmp - pp.PastProfit) / hours)
                    {
                        if (tmp != 0.0)
                        {
                            profit += tmp;
                            RemoveProfits(i);
                            pp.EmptyPastProfits();
                            ClosedAmount++;
                        }
                    }
                }
                if (i > 0)
                {
                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                            profit += (lp.Profit);
                        }
                    }
                }
            }
            return profit;
        }

        public double CalculateProfitWithProfit(double profitMin, double profitMax, out int ClosedAmountMin, out int ClosedAmountMax, out int RemovedMinAm, out int RemovedMaxAm)
        {
            double profit = 0.0;
            double tmp;
            ClosedAmountMin = 0;
            ClosedAmountMax = 0;
            RemovedMinAm = 0;
            RemovedMaxAm = 0;

            RemovedProfits.Clear();

            for (int i = 0; i < LooksAmount(); i++)
            {
                tmp = ProfitAtLook(i);
                if (tmp <= profitMin)
                {
                    if (tmp != 0.0)
                    {
                        profit += tmp;
                        RemovedMinAm += RemoveProfits(i);
                        ClosedAmountMin++;
                    }
                }
                if (tmp >= profitMax)
                {
                    if (tmp != 0.0)
                    {
                        profit += tmp;
                        RemovedMaxAm += RemoveProfits(i);
                        ClosedAmountMax++;
                    }
                }
                if (i > 0)
                {
                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                            profit += (lp.Profit);
                        }
                    }
                }
            }
            return profit;
        }

        private void ReverseProfits(int nr)
        {
            foreach (LookProfit lp in Looks[nr].Profits)
            {
                if (!ReversedProfits.ContainsValue(lp.ProfitId))
                {
                    ReversedProfits.Add(lp.ProfitId, lp.ProfitId);
                }
            }
        }

        public double CalculateProfitWithReversingProfit(double profitMin, out int ClosedAmountMin, out int RemovedMinAm)
        {
            double profit = 0.0;
            double tmp;
            bool reverse = false;
            ClosedAmountMin = 0;
            RemovedMinAm = 0;

            RemovedProfits.Clear();
            ReversedProfits.Clear();

            for (int i = 0; i < LooksAmount(); i++)
            {
                if (reverse)
                {
                    ReverseProfits(i);
                }
                tmp = ProfitAtLook(i);
                if (tmp <= profitMin)
                {
                    profit += tmp;
                    RemovedMinAm += RemoveProfits(i);
                    ClosedAmountMin++;
                    reverse = !reverse;


                }
                if (i > 0)
                {
                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                            if (ReversedProfits.ContainsValue(lp.ProfitId))
                            {
                                profit -= (lp.Profit);
                            }
                            else
                            {
                                profit += (lp.Profit);
                            }
                        }
                    }
                }
            }
            return profit;
        }

        public double CalculateProfitWithProfitAndTrailing(double profitMin, double profitMax, out int ClosedAmountMin, out int ClosedAmountMax, out int RemovedMinAm, out int RemovedMaxAm)
        {
            const double FALLBACK = 0.0;
            double profit = 0.0;
            double tmp;
            double highestTmp = 0.0;
            bool found = false;
            ClosedAmountMin = 0;
            ClosedAmountMax = 0;
            RemovedMinAm = 0;
            RemovedMaxAm = 0;


            RemovedProfits.Clear();

            for (int i = 0; i < LooksAmount(); i++)
            {
                tmp = ProfitAtLook(i);
                if (tmp <= profitMin)
                {
                    if (tmp != 0.0)
                    {
                        profit += tmp;
                        RemovedMinAm += RemoveProfits(i);
                        ClosedAmountMin++;
                    }
                }
                if (!found)
                {
                    if (tmp >= profitMax)
                    {
                        found = true;
                        highestTmp = tmp;
                    }
                }
                else
                {
                    if (tmp <= highestTmp - FALLBACK)
                    {
                        profit += tmp;
                        RemovedMaxAm += RemoveProfits(i);
                        ClosedAmountMax++;
                        found = false;
                    }
                }
                if (i > 0)
                {
                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                                profit += (lp.Profit);
                        }
                    }
                }
                highestTmp = tmp > highestTmp ? tmp : highestTmp;
            }
            return profit;
        }

        public double CalculateBalance(out double profitDiffrenceSum)
        {
            double profit = 0.0;

            profitDiffrenceSum = 0.0;
            RemovedProfits.Clear();

            for (int i = 0; i < LooksAmount(); i++)
            {
                if (i > 0)
                {
                    bool diff = false;

                    foreach (LookProfit lp in Looks[i - 1].Profits)
                    {
                        if (ClosedProfit(i, lp.ProfitId))
                        {
                            profit += lp.Profit;
                            diff = true;
                        }
                    }
                    if (diff)
                    {
                        profitDiffrenceSum += (Looks[i - 1].Profit - Looks[i].Profit);
                    }
                }
            }
            return profit;
        }

        public void MinMxProfits(out double min, out double max)
        {
            min = Looks[0].Profit;
            max = min;

            foreach (Look l in Looks)
            {
                if (l.Profit > max)
                {
                    max = l.Profit;
                }
                else if (l.Profit < min)
                {
                    min = l.Profit;
                }
            }
        }

        public void ReverseProfits()
        {
            foreach (Look l in Looks)
            {
                l.RevereProfits();
            }
        }

        public void NormalizeProfits()
        {
            foreach (Look l in Looks)
            {
                l.NormalizeProfits();
            }
        }

        public void DisplayProfitConsistency()
        {
            foreach (Look l in Looks)
            {
                if (l.ConsisitentProfit())
                {
                    Console.Write(" ok");
                }
                else
                {
                    l.ConsisitentProfit();
                    Console.WriteLine(" error");
                }
            }
        }
    }
}
