﻿/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Mail;
using System.Reflection.Emit;
using System.Text.RegularExpressions;
using System.Xml;

namespace CommonLibrary
{
    public class Profit : IComparable
    {
        public long Nr { get; set; }
        public bool Ended { get; set; } = false;
        public bool Successfull { get; set; } = false;
        public DateTime TradeBegin { get; set; }
        public DateTime TradeEnd { get; set; }
        public bool AfterTrendReverseTrades { get; set; } = false;
        public string Symbol { get; set; }
        public double GainLoss { get; set; }
        public bool Buy { get; set; }
        public double LotSize { get; set; }

        public double Price { get; set; }
        public bool Reversed { get; set; }

        public int CompareTo(object obj)
        {
            Profit other = obj as Profit;

            return Nr.CompareTo(other.Nr);
        }

        public static int CompareTradeBegin(Profit a, Profit b)
        {
            return a.TradeBegin.CompareTo(b.TradeBegin);
        }

        public static List<Profit> ParseMQL4File(string name)
        {
            StreamReader reader = new StreamReader(name);
            List<Profit> profits = new List<Profit>();
            string line;

            while (((line = reader.ReadLine()) != null) && (!line.Contains("<tr align=right>"))) ;

            if (line == null)
            {
                return null;
            }
            else
            {
                while (((line = reader.ReadLine()) != ""))
                {
                    Profit profit = GetProfit(line);

                    profits.Add(profit);
                }
            }
            return profits;
        }

        private static Profit GetProfit(string line)
        {
            if (line.Contains("bgcolor"))
            {
                line = line.Substring(20);
            }
            else
            {
                line = line.Substring(4);
            }

            string resWords = Regex.Replace(line, @"[^a-z]", " ");
            string[] arrWords = resWords.Split(new char[] { ' ' });
            string resNumbers = Regex.Replace(line, @"[^0-9.-]", " ");
            string[] arrNumbers = resNumbers.Split(new char[] { ' ' });

            Profit profit = new Profit();

            profit.TradeBegin = DateTime.Parse(GetPosNr(arrNumbers, 2) + " " + GetPosNr(arrNumbers, 3) + ":" + GetPosNr(arrNumbers, 4) + ":" + GetPosNr(arrNumbers, 5));
            if (line.Contains("buy"))
            {
                profit.Buy = true;
            }
            else
            {
                profit.Buy = false;
            }
            profit.LotSize = double.Parse(GetPosNr(arrNumbers, 6));
            profit.Price = double.Parse(GetPosNr(arrNumbers, 11));
            profit.Symbol = GetFirstLength(arrWords, 6);
            profit.TradeEnd = DateTime.Parse(GetPosNr(arrNumbers, 22) + " " + GetPosNr(arrNumbers, 23) + ":" + GetPosNr(arrNumbers, 24) + ":" + GetPosNr(arrNumbers, 25));
            profit.GainLoss = double.Parse(GetLastOne(arrNumbers));
            return profit;
        }

        private static string GetLastOne(string[] arr)
        {
            for (int i = arr.Length - 1; i >= 0; i--)
            {
                if (arr[i] != "")
                {
                    return arr[i];
                }
            }
            return null;
        }

        private static string GetFirstLength(string[] arr, int len)
        {
            int found = 0;

            foreach (string str in arr)
            {
                if (str.Length == len)
                {
                    if(found < 2)
                    {
                        found++;
                    }
                    else
                    {
                        return str;
                    }
                }
            }
            return null;
        }

        private static string GetPosNr(string[] arr, int n)
        {
            int pos = 0;

            for (int i = 0; i < arr.Length; i++)
            {
                if (pos == n)
                {
                    return arr[i - 1];
                }
                else
                {
                    if (arr[i] != "")
                    {
                        pos++;
                    }
                }
            }
            return null;
        }

        public static void Normalize(List<Profit> profits)
        {
            foreach (Profit profit in profits)
            {
                profit.GainLoss /= (profit.LotSize / 0.01);
            }
        }

        public static void SetReverseTrades(List<Profit> profits, DateTime date)
        {
            foreach (Profit profit in profits)
            {
                if (profit.TradeEnd >= date)
                {
                    profit.GainLoss = -profit.GainLoss;
                }
            }
        }

        public static bool AreEqual(Profit a, Profit b)
        {
            if(a.AfterTrendReverseTrades != b.AfterTrendReverseTrades ||
               a.Buy != b.Buy ||
               a.Ended != b.Ended ||
               a.GainLoss != b.GainLoss ||
               a.LotSize != b.LotSize ||
               a.Price != b.Price ||
               a.Successfull != b.Successfull ||
               a.Symbol != b.Symbol ||
               a.TradeBegin != b.TradeBegin ||
               a.TradeEnd != b.TradeEnd)
            {
                return false;
            }
            else
            {
                return true;
            }
        }
    }
}
