﻿/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using CommonLibrary;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;

namespace MethodGCalculator
{
    public class ComparerTradeBegin : IComparer<Profit>
    {
        public int Compare(Profit x, Profit y)
        {
            return (Profit.CompareTradeBegin(x, y));
        }
    }

    public class ComparerTradeEnd : IComparer<Profit>
    {
        public int Compare(Profit x, Profit y)
        {
            return x.TradeEnd.CompareTo(y.TradeEnd);
        }
    }

    public class Reverse
    {
        public bool ReverseState { get; set; }

        public DateTime Time { get; set; }
    }
    public class ProfitsModifier
    {
        public List<Profit> Profits { get; set; }

        private List<Reverse> Reverses { get; set; }
        private double AverageGain { get; set; }
        private double AverageLoss { get; set; }
        private void Sort()
        {
            Profits.Sort(new ComparerTradeBegin());
        }
        private void NormalizeLotSizes()
        {
            foreach (Profit p in Profits)
            {
                if (p.LotSize != 0.01)
                {
                    int am = (int)(p.LotSize / 0.01);

                    p.GainLoss /= am;
                    p.LotSize = 0.01;
                }
            }
        }

        private void CalculateAverages()
        {
            int amGain, amLoss;

            amGain = 0;
            amLoss = 0;
            AverageGain = 0.0;
            AverageLoss = 0.0;

            foreach (Profit p in Profits)
            {
                if (p.GainLoss >= 0.0)
                {
                    AverageGain += p.GainLoss;
                    amGain++;
                }
                else
                {
                    AverageLoss += p.GainLoss;
                    amLoss++;
                }
            }

            if (amGain > 0)
            {
                AverageGain /= amGain;
            }

            if (amLoss > 0)
            {
                AverageLoss /= amLoss;
            }
        }

        private void ReadReverses(string name)
        {
            Reverses = new List<Reverse>();
            StreamReader reader = new StreamReader(name);
            string line;

            while ((line = reader.ReadLine()) != "" && line != null)
            {
                string[] arr = line.Split(';');
                Reverse r = new Reverse();

                r.Time = DateTime.ParseExact(arr[0], "dd.MM.yyyy HH:mm", CultureInfo.InvariantCulture);

                if (arr[1] == "true")
                {
                    r.ReverseState = true;
                }
                else
                {
                    r.ReverseState = false;
                }
                Reverses.Add(r);
            }
        }

        private void ApplyReverses()
        {
            foreach (Profit p in Profits)
            {
                if (IsReversed(p.TradeBegin))
                {
                    p.Reversed = true;
                    p.GainLoss = -p.GainLoss;
                    p.Buy = !p.Buy;
                }
                else
                {
                    p.Reversed = false;
                }
            }
        }

        private void NormalizeReverses()
        {
            double percent = (-AverageLoss - AverageGain) / AverageGain;

            foreach (Profit p in Profits)
            {
                if (p.Reversed)
                {
                    if (p.GainLoss >= 0)
                    {
                        p.GainLoss -= percent * p.GainLoss;
                    }
                    else
                    {
                        p.GainLoss += percent * p.GainLoss;
                    }
                }
            }
        }

        public void Prepare(string name, string reverseName)
        {
            Profits = CommonLibrary.Profit.ParseMQL4File(name);
            Sort();
            NormalizeLotSizes();
            CalculateAverages();
            ReadReverses(reverseName);
            ApplyReverses();
            NormalizeReverses();
        }


        private List<Profit> Read(string name)
        {
            List<Profit> profits = new List<Profit>();
            StreamReader reader = new StreamReader(name);
            string line;
            string[] arr, res;

            while ((line = reader.ReadLine()) != "" && line != null)
            {
                Profit profit = new Profit();
                arr = line.Split(',');

                res = arr[0].Split(';');
                profit.Symbol = res[1];

                res = arr[1].Split(';');
                profit.Buy = ((res[1]) == "true");

                res = arr[6].Split(';');
                profit.LotSize = double.Parse(res[1]);

                res = arr[7].Split(';');
                profit.GainLoss = double.Parse(res[1]);

                res = arr[8].Split(';');
                profit.TradeBegin = GetDate(res[1]);

                res = arr[9].Split(';');
                profit.TradeEnd = GetDate(res[1]);

                profits.Add(profit);
            }

            return profits;
        }

        private DateTime GetDate(string str)
        {
            bool contains = false;

            if (!str.Contains("PM") && !str.Contains("AM"))
            {
                foreach (char c in str)
                {
                    if (c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z')
                    {
                        contains = true;
                        break;
                    }
                }
            }
            if(contains)
            {
                str.Replace("Jan", "01");
                str.Replace("Feb", "02");
                str.Replace("Mar", "03");
                str.Replace("Apr", "04");
                str.Replace("May", "05");
                str.Replace("June", "06");
                str.Replace("Jun", "06");
                str.Replace("July", "07");
                str.Replace("Jul", "07");
                str.Replace("Aug", "08");
                str.Replace("Sept", "09");
                str.Replace("Sep", "09");
                str.Replace("Oct", "10");
                str.Replace("Nov", "11");
                str.Replace("Dec", "12");

                return DateTime.ParseExact(str.Substring(4), "MM dd HH:mm:ss yyyy", CultureInfo.InvariantCulture);
            }
            return DateTime.Parse(str);
        }

        public void Prepare(string name)
        {
            Profits = Read(name);
            Sort();
            NormalizeLotSizes();
        }

        private bool IsReversed(DateTime time)
        {
            bool reversed = false;

            foreach (Reverse r in Reverses)
            {
                if (r.Time <= time)
                {
                    if (r.ReverseState)
                    {
                        reversed = true;
                    }
                    else
                    {
                        reversed = false;
                    }
                }
                else
                {
                    break;
                }
            }
            return reversed;
        }

        public void Save(string name)
        {
            Profits.Sort(new ComparerTradeEnd());
            StreamWriter writer = new StreamWriter(name);
            double sum = 0.0;

            foreach (Profit p in Profits)
            {
                sum += p.GainLoss;
                writer.WriteLine(p.TradeEnd.ToString() + ";" + sum.ToString());
            }
            writer.Flush();
            writer.Close();
        }

        public void SaveAsTradesClosed(string name)
        {
            string line;
            Profits.Sort(new ComparerTradeEnd());
            StreamWriter writer = new StreamWriter(name);

            foreach (Profit p in Profits)
            {
                line = "Symbol;" + p.Symbol + ", Buy;" + ((p.Buy) ? "true" : "false") + ", Price;" + p.Price + ", Cost;" + p.Price + ", TakeProfit;" + 0.0 + ", Stop Loss;" + 0.0 + ", Lot Size;" + p.LotSize + ", Profit;" + p.GainLoss + ", Open Time;" + p.TradeBegin + ", Close Time;" + p.TradeEnd;
                writer.WriteLine(line);
            }
            writer.Flush();
            writer.Close();

        }
    }
}
