﻿/*Copyright 2024 Inconsoft Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using CommonLibrary;
using System;
using System.Collections.Generic;
using System.Net;

namespace MethodGCalculator
{
    public class ProfitsRunner
    {
        public List<Profit> Profits { private get; set; }

        private List<Profit> OpenTrades { get; set; }
        
        private double Hours { get; set; }

        private bool ReverseTrades { get; set; }

        public ProfitsRunner(double hours)
        {
            OpenTrades = new List<Profit>();
            Hours = hours;
        }
        
        private void ReverseProfit(Profit profit)
        {
            profit.GainLoss = -profit.GainLoss;
        }

        private void AddOpenTrade(Balances Balance, Profit profit)
        {
            double sum = 0.0;
            List<Profit> rem = new List<Profit>();

            if(ReverseTrades)
            {
                ReverseProfit(profit);
            }

            OpenTrades.Add(profit);
            foreach(Profit p in OpenTrades)
            {
                if(p.TradeEnd <= profit.TradeBegin)
                {
                    sum += p.GainLoss;
                    rem.Add(p);
                }
            }
            Balance.AddBalance(profit.TradeBegin, sum);            
        }

        private void RemoveFromOpenTrades(List<Profit> profits)
        {
            foreach(Profit p in profits)
            {
                OpenTrades.Remove(p);
            }
        }

        private void CloseOpenTrades(Balances Balance, DateTime time)
        {
            double sum = 0.0;

            foreach (Profit p in OpenTrades)
            {
                sum += CalculateProfit(p, time);
            }
            Balance.AddBalance(time, sum);
            OpenTrades.Clear();
        }

        public List<SpeedInfo> Run(int amSpeeds)
        {
            List<SpeedInfo> speeds = new List<SpeedInfo>();
            double minSpeed = 0.0, maxSpeed = 0.0;
            double diff;
            Balances Balance;

            CalculateMinMaxSpeeds(ref minSpeed, ref maxSpeed);
            diff = (maxSpeed - minSpeed) / amSpeeds;
            OpenTrades.Clear();

            for (double speed = minSpeed; speed <= maxSpeed; speed += diff)
            {
                Balance = new Balances(Hours);
                ReverseTrades = false;

                Console.Write("{0}/{1}. ", speed, maxSpeed);

                foreach (Profit profit in Profits)
                {
                    AddOpenTrade(Balance, profit);
                    if(Balance.IsSpeedValid())
                    {
                        if(Balance.Speed() <= speed)
                        {
                            CloseOpenTrades(Balance, profit.TradeBegin);
                            ReverseTrades = !ReverseTrades;
                            Balance.SetReset(profit.TradeBegin);
                        }
                    }
                }
                SpeedInfo si = new SpeedInfo();

                si.Profit = Balance.FinalBalance();
                si.Speed = speed;
                si.Hours = Hours;

                speeds.Add(si);
            }
            return speeds;
        }

        private void CalculateMinMaxSpeeds(ref double minSpeed, ref double maxSpeed)
        {
            Balances Balance = new Balances(Hours);
            minSpeed = 0.0;
            maxSpeed = 0.0;

            foreach(Profit profit in Profits)
            {
                AddOpenTrade(Balance, profit);
                if (Balance.IsSpeedValid())
                {
                    double speed = Balance.Speed();

                    if(speed > maxSpeed)
                    {
                        maxSpeed = speed;
                    }
                    if(speed < minSpeed)
                    {
                        minSpeed = speed;
                    }
                }
            }
        }

        private double CalculateProfit(Profit profit, DateTime time)
        {
            double distBegin = DateTimeHelper.DifferenceHours(profit.TradeBegin, time);
            double distEnd = DateTimeHelper.DifferenceHours(time, profit.TradeEnd);
            double endWeight = distBegin / (distBegin + distEnd);

            return endWeight * profit.GainLoss;
        }
    }
}